#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
__author__ = 'You Lu <you.lu@ukri.stfc.org>'
def getDictEntryOLD(dictObj, key1, key2=None):
    if key2:
        try:
            return dictObj[key1][key2]
        except:
            return None
    else:
        try:
            return dictObj[key1]
        except:
            return None
def getDictEntry(dictObj, *args):
    buff = dictObj
    for key in args:
        try:
            buff = buff.get(key)
            if buff == None:
                raise AttributeError
        except AttributeError:
            for val in list(dictObj.values()):
                try:
                    buff = getDictEntry(val, *args)
                    if buff == None:
                        continue
                    return buff
                except:
                    continue
    return buff
def hasDictEntry(dictObj, key1, key2=None):
    if key2:
        try:
            dictObj[key1][key2]
            return True
        except:
            return False
    else:
        try:
            dictObj[key1]
            return True
        except:
            return False
def getSortedKeys(dictObj, order=None):
    if order:
        return   [ key for key in order if key in dictObj ] \
               + [ key for key in dictObj if key not in order ]
    else:
        return sorted(list(dictObj.keys()))
def dict2obj(d, obj):
    from . import objutils
    for key in obj._kwargs.keys():
        setattr(obj, key, d[key])
def getMatchedKey(d, arg):
    l = [ key for key in d.keys() if arg in key ]
    if len(l) == 1:
        return l[0]
    elif len(l) > 1:
        return l
    else:
        return arg
def getValue(d, key, case_sensitive=True, raiseError=True):
    try:
        key = key.decode()
    except:
        pass
    try:
        if not case_sensitive:
            mapping = { k.lower():v for k,v in d.items() }
            return mapping[key.lower()]
        else:
            return mapping[key]
    except:
        if raiseError:
            raise
        else:
            return None
def isKeyOf(d, key, case_sensitive=True):
    from numpy import array, char
    try:
        key = key.decode()
    except:
        pass
    if not case_sensitive:
        keys = char.lower(array(list(d.keys())))
        return key.lower() in keys
    else:
        keys = array(list(d.keys()))
        return key in keys
def mergeDicts(d1, d2):
    from numpy import array, append
    for k, v in d2.items():
        if type(v) is dict:
            dbuff = d1.get(k, {})
            mergeDicts(dbuff, v)
            d1.update({k:dbuff})
        elif type(v) in [ str, float, int, bool ]:
            pass
        else:
            dt = v.dtype
            d1.update({k:append(d1.get(k, array([], dtype=dt)), v)})



#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
__author__ = 'You Lu <you.lu@ukri.stfc.org>'
from ctypes import CDLL, RTLD_LOCAL, RTLD_GLOBAL
from os     import path
def importModule(modname, filepath, relative_dir='.', suppress_warning=True, submodules=None, **kwargs):
    from importlib.util import spec_from_file_location, module_from_spec
    from warnings       import catch_warnings, resetwarnings, simplefilter
    from os.path        import abspath, dirname, join
    from sys            import path, stderr, stdout
    pkgdir = abspath(join(dirname(filepath), relative_dir))
    if pkgdir not in path:
        path.append(pkgdir)
    spec   = spec_from_file_location(modname, filepath, submodule_search_locations=submodules)
    module = module_from_spec(spec)
    for k, v in kwargs.items():
        setattr(module, k, v)
    if suppress_warning:
        with catch_warnings(record=True) as w:
            simplefilter('ignore')
            spec.loader.exec_module(module)
        resetwarnings()
    else:
        spec.loader.exec_module(module)
    stderr.flush()
    stdout.flush()
    return module
def getSharedLibname(modname):
    return 'lib' + modname.lower() + '.so'
def isLinked(modname, location):
    lib = getSharedLib(modname, location)
    if modname in [ 'parallel' ]:
        return True
    try:
        return bool(lib.islinked())
    except:
        return False
def getSharedLib(modname, location, mode=RTLD_GLOBAL, debug=0):
    libname = getSharedLibname(modname)
    try:
        soname = path.join(location, libname.replace('-', '_'))
        return CDLL(soname, mode=mode)
    except OSError as e:
        if 'undefined symbol' in str(e):
            raise
        if debug > 5:
            print(" >>> dl_py2f.utils.modutils.getSharedLib({}) failed to initialise CDLL:\n".format(modname), e)
        try:
            soname = path.join(location, libname.replace('_', '-'))
            return CDLL(soname, mode=mode)
        except:
            return
def getLinkedLib(modname, location, mode=RTLD_GLOBAL, is_linked=False):
    lib = getSharedLib(modname, location, mode=mode)
    if isLinked(modname, location) or is_linked:
        return lib
    else:
        return



#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
__author__     = [ 'You Lu', 'Thomas W. Keal' ]
__copyright__  = 'Copyright (C) 2025 The authors of DL_PY2F'
__credits__    = [ 'You Lu', 'Thomas W. Keal' ]
__license__    = 'LGPLv3'
__version__    = '1.1.8'
__maintainer__ = __author__[0]
__email__      = 'you.lu@stfc.ac.uk'
__status__     = ''
__dictname = '_kwargs'
__arrname = '_master'
def __getType(obj):
    from numpy  import ndarray, float64, float32, int64, int32
    from types  import ModuleType, FunctionType
    if isinstance(obj, ndarray):
        return list
    elif isinstance(obj, list):
        return list
    elif isinstance(obj, str):
        return str
    elif isinstance(obj, bytes):
        return bytes
    elif isinstance(obj, float64) or isinstance(obj, float32):
        return float
    elif isinstance(obj, int64) or isinstance(obj, int32):
        return int
    elif isinstance(obj, ModuleType):
        return ModuleType
    elif isinstance(obj, FunctionType):
        return FunctionType
    elif hasattr(obj, __dictname) and type(obj).__module__ != 'builtins':
        return object
    else:
        return type(obj)
def py2f(obj, debug=0, byref=False):
    from ctypes import CFUNCTYPE, c_int, c_long, c_double, c_bool, c_char, c_wchar_p, c_char_p, c_void_p, addressof, pointer, POINTER, Structure
    from numpy  import array, asarray, ctypeslib, ma, str_
    from types  import ModuleType
    from sys    import stdout
    if debug > 2:
        print("\n DL_PY2F: obj type = {}\n".format(type(obj)))
        stdout.flush()
    MAXLEN = 256
    ATTRLEN = 32
    dictbuff = getattr(obj, __dictname)
    _kwargs = sorted(dictbuff.items())
    fbuff       = []
    initialiser = []
    def _appendInitAtoF(_initialiser, _obj, _key, _default_type):
        if isinstance(_obj, ma.core.MaskedArray):
            databuff = _obj.torecords()['_data'][_key]
        else:
            databuff = getattr(_obj, _key)
        try:
            _initialiser.append((__getType(databuff).__name__ + " "*ATTRLEN)[:ATTRLEN].encode('ascii'))
        except:
            _initialiser.append((__getType(databuff).__class__.__name__ + " "*ATTRLEN)[:ATTRLEN].encode('ascii'))
        try:
            _initialiser.append((_obj.synons[_key.strip()] + " "*ATTRLEN)[:ATTRLEN].encode('ascii'))
        except:
            _initialiser.append((_key+' '*ATTRLEN)[:ATTRLEN].encode('ascii'))
        if _default_type is list:
            abuff = asarray(databuff)
            if abuff.dtype.name.startswith('void'):
                size = abuff.size*len(abuff.dtype.names)
            else:
                size = abuff.size
            _initialiser += [ (abuff.dtype.name+' '*ATTRLEN)[:ATTRLEN].encode('ascii'),
                             c_long(size),
                             c_long(abuff.shape[0]) ]
        else:
            _initialiser += [ (' '*ATTRLEN)[:ATTRLEN].encode('ascii') ] + [ c_long(0) ]*2
        if isinstance(_obj, ma.core.MaskedArray):
            _initialiser.append(c_bool(_default_type is list and _key in _obj.torecords()['_data'].dtype.fields))
        else:
            _initialiser.append(c_bool(_default_type is list and _key in getattr(_obj, '_fields', [])))
    def __list2ndp(obj, key, foo):
        if hasattr(foo, "dtype"):
            selectcases = { 'int64'   : c_long,
                            'int32'   : c_int,
                            'float64' : c_double,
                            'float32' : c_double,
                            'bool'    : c_long,
                            'object'  : c_long,
                            'bytes64' :'U8',
                            'bytes32' :'U8',
                            'bytes128':'U8',
                            'str32'   :'U8',
                            'str64'   :'U8',
                            'str256'  :'U8',
                            'record'  : foo.dtype,
                            'void'    : foo.dtype,
                          }
            def _addNPArray(_ctype, _key, _val):
                npptr = ctypeslib.ndpointer(_ctype)
                if _val.size > 0:
                    npptr._shape_ = _val.shape
                fbuff.append((_key, npptr))
                initialiser.append(_val.ctypes.data)
            def _addRecArray(_ctype, _arr):
                for field in _ctype.names:
                    _appendInitAtoF(initialiser, foo, field, list)
                    _addNPArray(selectcases[_ctype.fields[field][0].name], field, _arr[field])
            def _addFlexible(_ctype, _key, _val):
                npptr = ctypeslib.ndpointer(_ctype)
                if _val.size > 0:
                    npptr._shape_ = (_val.shape[0],len(_val.dtype.names))
                fbuff.append((_key, npptr))
                initialiser.append(_val.ctypes.data)
            try:
                _addNPArray(selectcases[foo.dtype.name], key, getattr(obj, key))
            except KeyError:
                if foo.dtype.name[:6] == 'record':
                    _addRecArray(selectcases[foo.dtype.name[:6]], getattr(obj, key))
                elif foo.dtype.name[:4] == 'void':
                    _addFlexible(selectcases[foo.dtype.name[:4]], key, getattr(obj, key))
        else:
            try:
                initialiser.append(asarray([i for i in getattr(obj, key)], dtype=c_long).ctypes.data)
                fbuff.append((key, ctypeslib.ndpointer(c_long)))
            except ValueError:
                if all([type(i) in [str,str_] for i in getattr(obj, key)]):
                    fbuff.append((key, c_char_p))
                    initialiser[-6] = ('str'+' '*ATTRLEN)[:ATTRLEN].encode('ascii')
                    initialiser.append(c_char_p((';'.join(foo) + " "*MAXLEN)[:MAXLEN].encode('ascii')))
                else:
                    initialiser.append(None)
                    fbuff.append((key, c_void_p))
            except:
                if debug > 2:
                    print(" >>> DL_PY2F WARNING: cannot convert non-integer list/tuple `%s` of"%key, obj, "to NumPy ndarray")
                initialiser.append(None)
                fbuff.append((key, c_void_p))
    def __int2cint(obj, key, foo):
        fbuff.append((key, POINTER(c_long)))
        initialiser.append(pointer(c_long(foo)))
    def __func2cfunptr(obj, key, foo):
        from typing import get_type_hints
        selectcases = { float: c_double,
                        int  : c_long }
        c_type = selectcases.get(get_type_hints(foo).get('return', int), c_long)
        fbuff.append((key, POINTER(CFUNCTYPE(c_type, c_void_p))))
        initialiser.append(pointer(CFUNCTYPE(c_type, c_void_p)(foo)))
    def __str2bytes(obj, key, foo):
        fbuff.append((key, c_char_p))
        initialiser.append(c_char_p((foo + " "*MAXLEN)[:MAXLEN].encode('ascii')))
    def __module2bytes(obj, key, foo):
        fbuff.append((key, c_char_p))
        initialiser.append(c_char_p((foo.__name__ + " "*MAXLEN)[:MAXLEN].encode('ascii')))
    def __obj2ptr(obj, key, foo):
        bar = py2f(foo)
        fbuff.append((key, POINTER(type(bar))))
        initialiser.append(pointer(bar))
    def __dict2None(obj, key, foo):
        if debug > 2:
            print(" >>> DL_PY2F WARNING: unsupport <class dict> entry \""+key+"\" will be treated as None.")
        fbuff.append((key, c_void_p))
        initialiser.append(None)
    def __float2cdouble(obj, key, foo):
        fbuff.append((key, POINTER(c_double)))
        initialiser.append(pointer(c_double(foo)))
    def __bool2cbool(obj, key, foo):
        fbuff.append((key, POINTER(c_bool)))
        initialiser.append(pointer(c_bool(foo)))
    def __None2ptr(obj, key, foo):
        fbuff.append((key, c_void_p))
        initialiser.append(None)
    selectcases = { 
                    bool            :__bool2cbool,
                    dict            :__dict2None,
                    float           :__float2cdouble,
                    int             :__int2cint,
                    list            :__list2ndp,
                    object          :__obj2ptr,
                    tuple           :__list2ndp,
                    type(lambda x:x):__func2cfunptr,
                    str             :__str2bytes,
                    ModuleType      :__module2bytes,
                    type(None)      :__None2ptr }
    for key, default_val in _kwargs:
        if key.startswith('_'):
            continue
        try:
            isRec = getattr(obj, key).dtype.name.startswith('record')
        except:
            isRec = False
        if not isRec:
            _appendInitAtoF(initialiser, obj, key, __getType(default_val))
        try:
            selectcases[__getType(getattr(obj, key))](obj, key, getattr(obj, key))
        except KeyError:
            print(" >>> DL_PY2F ERROR: type "+str(__getType(getattr(obj, key)))+" of entry \""+key+"\" not supported.\n")
    fields = [ item for sublist in zip(zip([ 'type'    + f[0]   for f in fbuff ],
                                           [ c_char*ATTRLEN     for f in fbuff ]),
                                       zip([ f[0] + ' '*ATTRLEN for f in fbuff ],
                                           [ c_char*ATTRLEN     for f in fbuff ]),
                                       zip([ 'dtype'   + f[0]   for f in fbuff ],
                                           [ c_char*ATTRLEN     for f in fbuff ]),
                                       zip([ 'size'    + f[0]   for f in fbuff ],
                                           [ c_long             for f in fbuff ]),
                                       zip([ 'sizem'   + f[0]   for f in fbuff ],
                                           [ c_long             for f in fbuff ]),
                                       zip([ 'isfield' + f[0]   for f in fbuff ],
                                           [ c_bool             for f in fbuff ]),
                                       fbuff)
                    for item in sublist ]
    if len(fields) != len(initialiser) or debug > 4:
        if len(fields) != len(initialiser):
            print(" >>> DL_PY2F WARNING: fields (len: {}) and initialiser (len: {}) mismatch:".format(len(fields),len(initialiser)))
            print("    Python object:", obj)
        else:
            print(" >>> DL_PY2F DEBUG: data structure of PY2F object")
        print(" "*4+"-"*48)
        print("    {:20}  |  {}".format('field', 'initialiser'))
        for i in range(max(len(fields), len(initialiser))):
            try:
                if type(initialiser[i]) is bytes and fields[i][0].startswith('type'):
                    print(" "*4+"-"*48)
                print("    {:20}  |  {}".format(fields[i][0].strip(), initialiser[i]))
            except IndexError:
                try:
                    print("    {:20}  |  {}".format('', initialiser[i]))
                except:
                    print("    {:20}  |    ".format(fields[i][0].strip()))
        print(" "*4+"-"*48)
    stdout.flush()
    fields.insert(0, ('nattrs', c_long))
    initialiser.insert(0, c_long(int((len(fields)-1)/7)))
    fields.insert(1, ('_class', c_char_p))
    initialiser.insert(1, c_char_p((obj.__module__ + '.' + obj.__class__.__name__ + " "*MAXLEN)[:MAXLEN].encode('ascii')))
    fields.insert(2, ('_address', c_long))
    initialiser.insert(2, addressof(obj))
    fields.insert(3, ('width', c_long))
    try:
        if hasattr(obj, __arrname):
            _master = getattr(obj, __arrname)
            width_list = [ _master[field].shape[1] if _master[field].ndim == 2 else
                           array(_master.dtype.fields[field][0].names).size for field in _master.dtype.fields ]
            _width = sum(width_list)
        initialiser.insert(3, c_long(_width))
    except:
        initialiser.insert(3, c_long(0))
    fields.insert(4, ('debug', POINTER(c_long)))
    initialiser.insert(4, pointer(c_long(debug)))
    stdout.flush()
    if debug >= 5:
        print("\n >>> DL_PY2F DEBUG: Components of wrapped Python object", obj)
        print("\n     {:5}    {:32}    {:24} {}".format('index', 'field name', 'field type', 'value'))
        for i, v in enumerate(initialiser):
            if not (i-5)%7:
                print()
            print("     {:5d}    {:32.32}    {:24.24} {}".format(i, '\"'+fields[i][0]+'\"', fields[i][1].__name__, v))
    class CStruct(Structure):
        _fields_ = fields
        def __new__(cls, *args):
            inst = Structure.__new__(cls)
            stdout.flush()
            return inst
    if not byref:
        return CStruct(*initialiser)
    else:
        from ctypes import byref as ctypes_byref
        from sys    import stdout
        try:
            return ctypes_byref(CStruct(*initialiser))
        except:
            print(" >>> DL_PY2F ERROR: obj =", obj)
            stdout.flush()
            raise
import os, sys
sys.path.append(os.path.dirname(__file__))
from .dl_f2py import DL_DL
from . import utils



#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
__author__ = 'You Lu <you.lu@ukri.stfc.org>'
from . import miscutils
def getFlattenedIter(iter, sort=True):
    from itertools   import chain
    from collections import abc, OrderedDict
    if not hasattr(iter, '__len__'):
        iter = [ iter ]
    l = [ [ j for j in i ] if isinstance(i, abc.Iterable) and not isinstance(i, str) else
          [ i ] for i in iter ]
    if sort:
        return list(set(chain.from_iterable(l)))
    else:
        return list(OrderedDict.fromkeys(chain.from_iterable(l)))
def getFlattened(iter, unique=True):
    from numpy    import array, dtype, setdiff1d, __version__
    from warnings import filterwarnings
    from numpy.lib import NumpyVersion
    if NumpyVersion(__version__) >= '2.0.0b1':
        from numpy.exceptions import VisibleDeprecationWarning
    else:
        from numpy import VisibleDeprecationWarning
    if type(iter) in [dict]:
        return iter
    try:
        with filterwarnings('error', category=VisibleDeprecationWarning) as filter:
            try:
                abuff = array(iter)
            except:
                abuff = array(iter, dtype=object)
        if abuff.dtype != dtype('O'):
            return abuff.ravel().tolist()
    except:
        pass
    filterwarnings('error')
    try:
        flattened = type(iter)()
        type_iter = type(iter)
    except:
        flattened = list()
        type_iter = list
    for item in iter:
        if type(item) is type_iter:
            tmp = getFlattened(item, unique=unique)
            for t in tmp:
                if t in flattened and unique:
                    flattened.remove(t)
                flattened += type_iter([t])
        else:
            if unique:
                try:
                    item = setdiff1d(item, flattened)
                except RuntimeWarning:
                    item = [item]
                try:
                    flattened += type_iter(type_iter(item))
                except:
                    flattened += type_iter([item])
            elif not unique:
                try:
                    flattened += type_iter(type_iter(item))
                except:
                    flattened += type_iter([item])
    filterwarnings('default')
    return flattened
def getFlattenedOLD(iter, unique=True):
    flattened = type(iter)()
    for item in iter:
        if type(item) is type(iter):
            tmp = getFlattened(item)
            for t in tmp:
                if t in flattened:
                    flattened.remove(t)
                flattened += type(iter)([t])
        else:
            if unique and item not in flattened:
                flattened += type(iter)([item])
            elif not unique:
                flattened += type(iter)([item])
    return flattened
def findMapped(iter, mapping, case_sensitive=True):
    from numpy import array, char, isin
    iter = array(iter)
    try:
        if not case_sensitive:
            iter = char.lower(char.decode(iter))
        else:
            iter = char.decode(iter)
    except:
        if not case_sensitive:
            iter = char.lower(iter)
    if mapping:
        if not case_sensitive:
            mapping_original, mapping = mapping, { k.lower():v for k,v in mapping.items() }
            keys_original = array(list(mapping_original.keys()))
        else:
            keys_original = array(list(mapping.keys()))
        keys   = array(list(mapping.keys()))
        mapped = keys_original[isin(keys,iter)]
    else:
        mapped = array([])
    return mapped



#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
__author__ = 'You Lu <you.lu@ukri.stfc.org>'
def file2text(filename, case='input', error_return='', preceding_newline=False, trailing_newline=False, prefix=''):
    from os import path
    selectcases = {
                    'lower':str.lower,
                    'title':str.title,
                    'input':str
                  }
    filename0 = filename + ''
    filename = path.join(prefix, filename)
    try:
        with open(filename, 'r') as fp:
            strbuff = selectcases[case.lower()](fp.read())
        if preceding_newline:
            strbuff = '\n'+ strbuff
        if trailing_newline:
            strbuff = strbuff + '\n'
        return strbuff
    except (FileNotFoundError, IsADirectoryError, OSError):
        if error_return in [ False, None ]:
            raise FileNotFoundError(' >>> ERROR: file %s not found'%filename) from None
        elif error_return in [ '' ]:
            return filename0
        else:
            return error_return
def getBaseName(filename):
    from os.path import splitext
    return splitext(filename)[0]
def getExtension(filename, lower=True):
    from os import path
    filename = str(filename).lower()
    if path.splitext(filename)[1] != '':
        if lower:
            return path.splitext(filename)[1][1:].lower()
        else:
            return path.splitext(filename)[1][1:]
    else:
        return ''
def getExtensionWithDot(filename):
    from os.path import splitext
    if splitext(filename)[1]:
        return splitext(filename)[1]
    else:
        return None
def getOutName(filename):
    try:
        return getBaseName(filename) + ".out"
    except:
        raise IOError
def joinContinuedLines(fp, continuator='\\'):
    from os import SEEK_CUR
    fp.seek(fp.tell())
    line = fp.readline()
    while line.rstrip().endswith(continuator):
        line = ' '.join([line.rstrip().rstrip(continuator), fp.readline()])
    return line
def rename(filename, newname):
    from os import path, replace
    if path.isfile(filename):
        replace(filename, newname)
def archiveFile(filename, dir='.', copy=False, suffix='.OLD'):
    from os     import getcwd, listdir, path, sep
    from shutil import copyfile
    if type(filename) is tuple or type(filename) is list:
        for f in filename:
            archiveFile(f, dir=dir, suffix=suffix)
    elif type(filename) == str:
        if not path.isdir(dir):
            archiveDir(dir)
        dest = path.join(getcwd(), dir, filename.split(sep)[-1]+suffix)
        if copy:
            copyfile(filename, dest)
        else:
            rename(filename, dest)
def archiveDir(dirname, suffix='.OLD', mkdir=True):
    from shutil import rmtree
    from os     import mkdir as md
    from os     import replace
    rmtree(dirname+suffix, ignore_errors=True)
    try:
        replace(dirname, dirname+suffix)
    except FileNotFoundError:
        pass
    if mkdir:
        md(dirname)
def getChoppedMasks(filename, section_key, closing_key='', return_start_lines=False):
    from numpy import arange, array, append, char, searchsorted
    with open(filename, 'r') as fp:
        lines = array(fp.readlines())
        nlines = lines.size
        linenums = arange(nlines)
        linenums_mol = linenums[char.startswith(lines, section_key)]
        linenums_entries = linenums
        sorting_mol = searchsorted(linenums_entries, linenums_mol)
        lindices_mol = linenums_mol
        rindices_mol = linenums_entries[append(sorting_mol[1:]-1, linenums_entries.size-1)]
        masks = (lindices_mol[:,None]<=linenums)*(rindices_mol[:,None]>=linenums)
    if return_start_lines:
        return masks, lindices_mol
    else:
        return masks
def write(filename, buff):
    try:
        with open(filename, 'w') as fp:
            fp.write(buff)
    except:
        raise IOError



#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
__author__ = 'You Lu <you.lu@ukri.stfc.org>'
import numpy
from . import iterutils
def ndarray(buffer, shape=None, dtype=None):
    if shape is None:
        shape = numpy.array(buffer).shape
    if dtype is None:
        dtype = numpy.array(buffer).dtype
    arr = numpy.ndarray(shape  = shape,
                        dtype  = dtype,
                        buffer = numpy.array(buffer))
    return numpy.ndarray(shape  = shape,
                         dtype  = dtype,
                         buffer = numpy.array(buffer))
def intarray(buffer, shape=None):
    if shape is None:
        shape = numpy.array(buffer).shape
    return numpy.ndarray(shape  = shape,
                         dtype  = numpy.int64,
                         buffer = numpy.array(numpy.int64(buffer)))
def indexarray(shape=None):
    if shape is None:
        shape = (1)
    return numpy.full(shape,
                      -1,
                      dtype=numpy.int64,
                     )
def setField(a, buffer, debug=0):
    if numpy.array(buffer).size != a.size and debug > 4:
        print('\n >>> WARNING: dl_py2f.utils.nputils.setField() failed to reshape input data of shape {} to {}'.format(numpy.array(buffer).shape, a.shape))
    a.setfield(numpy.array(buffer).reshape(a.shape), a.dtype)
def inplaceAdd(a, buffer):
    a.__iadd__(numpy.array(buffer).reshape(a.shape))
def inplaceMultiply(a, buffer):
    a.__imul__(buffer)
def map(a):
    if a.dtype.name == 'object':
        mapped = numpy.ndarray(shape=(a.shape[0], a[0].size))
        for i in range(a.shape[0]):
            mapped[i].setfield(a[i][:], dtype=a[0][0].dtype)
        return mapped
class RecArray(numpy.recarray):
    @property
    def defaults(self):
        return self._defaults
    @defaults.setter
    def defaults(self, val):
        self._defaults = val
    def __new__(cls, shape=None, dtype=None, buff=None, **kwargs):
        try:
            dtype = numpy.dtype(dtype)
            inst = numpy.recarray.__new__(cls, shape, dtype=dtype, buf=buff, **kwargs)
        except:
            buff = [ i[1] if isinstance(i[0], tuple) else
                     None for i in dtype ]
            dtype = [ i[0] if isinstance(i[0], tuple) else
                      i    if isinstance(i   , tuple) else
                      None for i in dtype ]
            inst = numpy.recarray.__new__(cls, shape, dtype=dtype, **kwargs)
            inst.defaults = dict(zip(inst.dtype.names, buff))
            inst.init()
        return inst
    def init(self):
        if not self.shape[0]:
            return
        for key, val in self.defaults.items():
            if hasattr(val, '__len__'):
                self.setField(val, field=key)
            else:
                self.field(key, val)
    def expand(self, N, axis=0):
        N = int(N)
        if N == self.size:
            return
        size = self.size
        self.resize(N, refcheck=False)
        if size:
            self[size:] = self[0]
        else:
            self.init()
    def setField(self, buff, field=None, factor=1.0):
        def _fill(_buff, _field):
            if hasattr(_buff, '__len__'):
                try:
                    self[_field].setfield(numpy.array(_buff).reshape(self[_field].shape), self[_field].dtype)
                except:
                    self[_field][:] = numpy.array(_buff)
            else:
                self[_field].setfield(numpy.full_like(self[_field], _buff).reshape(self[_field].shape), self[_field].dtype)
            if factor != 1.0:
                self[_field].__imul__(factor)
        if field:
            _fill(buff, field)
        elif not field and buff.dtype.fields:
            self.expand(buff.shape[0])
            for field in buff.dtype.fields.keys():
                _fill(buff[field], field)
        elif not field and not buff.dtype.fields:
            try:
                self[:] = buff[:]
            except ValueError:
                self.expand(len(buff[:]))
                for i in range(len(self[:])):
                    self[i] = buff[i,:] 
    def getDtype(self, field):
        return getattr(self, field).dtype.descr
class SymArray(numpy.ndarray):
    def __setitem__(self, indices, val):
        lbuff = [ i for i in indices[:-1] if isinstance(i, (int, list, tuple, numpy.ndarray)) ][::-1]
        lbuff.append(indices[-1])
        reversed = tuple(lbuff)
        super(SymArray, self).__setitem__(indices , val)
        super(SymArray, self).__setitem__(reversed, val)
def dtype2fmt(dt):
    from numpy import dtype, typecodes
    from .     import dictutils
    selectcases = {
                    typecodes['AllInteger']: 'd',
                    typecodes['AllFloat']  : 'f',
                    typecodes['Character'] : 's',
                    'U':                     's',
                    'S':                     's',
                  }
    return selectcases[dictutils.getMatchedKey(selectcases, dtype(dt).char)]
def tolist(arr):
    return bytes2str(arr).tolist()
def bytes2str(arr):
    from numpy import str as np_str
    if arr.dtype.kind == 'S':
        return arr.astype(np_str, copy=False)
    else:
        return arr
def getCountsA2InA1(a1, a2):
    from numpy import arange, array, bincount, isin, searchsorted, unique, zeros
    a1, a2 = array(a1), array(a2)
    if len(a2) != unique(a2).size:
        print("\n WARNING: getCountsA2InA1() must not contain repeat! Returning zeros array...\n")
        return zeros((a2.size,), dtype=int)
    masks_a1_in_a2 = isin(a1, a2)
    a1_in_a2 = a1[masks_a1_in_a2]
    sorter = a2.argsort()
    indices = sorter[searchsorted(a2, a1_in_a2, sorter=sorter)]
    return bincount(indices)
def getCountsA1InA2(a1, a2):
    from numpy import arange, array, bincount, isin, searchsorted, unique, zeros
    a1, a2 = array(a1), array(a2)
    if len(a2) != unique(a2).size:
        print("\n WARNING: getCountsA1InA2() must not contain repeat! Returning zeros array...\n")
        return zeros((a1.size,), dtype=int)
    counts = zeros((a1.size), dtype=int)
    masks_a1_in_a2 = isin(a1, a2)
    a1_in_a2 = a1[masks_a1_in_a2]
    unique_a1_in_a2 = unique(a1_in_a2)
    indices = searchsorted(unique_a1_in_a2, a1_in_a2)
    abuff = bincount(indices)
    counts[masks_a1_in_a2] = abuff[indices]
    return counts
def getIncrementsFromCounts(arr):
    if not len(arr):
        return numpy.zeros(arr.shape, dtype=int)
    unique_a, counts_a = numpy.unique(arr, return_counts=True)
    increments = numpy.concatenate(tuple(numpy.arange(c) for c in counts_a))
    sorter = arr.argsort().argsort()
    return increments[sorter]
def getGroupedIndices(arr):
    from numpy import arange, argsort, cumsum, diff, split
    sorter = argsort(arr)
    arr_sorted = arr[sorter]
    masks_leading = [True] + (arr_sorted[1:] != arr_sorted[:-1]).tolist()
    increments = diff(arange(arr.shape[0])[masks_leading])
    return split(sorter, cumsum(increments))
def getDTypeFromObj(obj, dictname='_kwargs'):
    kwargs = getattr(obj, dictname)
    fields = [ obj._synons[field] if field in obj._synons else
               field for field in iterutils.getFlattened((obj._fields,)) ]
    dtypes = [ [field, kwargs[field].dtype] if kwargs[field].ndim == 1 else
               [field, kwargs[field].dtype, kwargs[field].shape[1]] if kwargs[field].ndim == 2 else
               [field, kwargs[field].dtype, 0]
               for field in iterutils.getFlattened((obj._fields,)) ]
    dtypes = [ tuple([obj._synons[dtype[0]]]+dtype[1:]) if dtype[0] in obj._synons else
               tuple(dtype) for dtype in dtypes ]
    values = [ kwargs[field] for field in iterutils.getFlattened((obj._fields,)) ]
    return list(zip(dtypes, values))



#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
__email__ = 'you.lu@stfc.ac.uk'
__author__ = f'You Lu <{__email__}>'
import os, utils
from importlib.resources import files
from ctypes    import addressof, CDLL, memset, RTLD_GLOBAL, sizeof
from ctypes    import c_bool, c_char_p, c_double, c_float, c_int, c_long, c_void_p, POINTER, Structure
from ctypes    import c_byte, c_char, c_short, c_size_t, c_uint, c_ubyte, c_ulong, c_void_p, c_wchar
from _ctypes   import _SimpleCData
from numpy     import array, dtype, empty, full, ndindex
from math      import ceil, prod
from itertools import product
from inspect   import stack
from weakref   import WeakValueDictionary
from time      import time
from .         import libpath
from .utils    import colours as _clr
_accessor_cache = {}
_DL_CACHE = WeakValueDictionary()
class c_complex8(Structure):
    _fields_ = [('re', c_double), ('im', c_double)]
class c_complex4(Structure):
    _fields_ = [('re', c_float), ('im', c_float)]
selectcases = { float      : c_double,
                c_double   : c_double,
                c_float    : c_float,
                int        : c_int,
                c_int      : c_int,
                c_long     : c_long,
                str        : c_char_p,
               'CHARACTER1': c_char_p,
               'COMPLEX4'  : c_complex4,
               'COMPLEX8'  : c_complex8,
               'INTEGER4'  : c_int,
               'INTEGER8'  : c_long,
               'REAL4'     : c_float,
               'REAL8'     : c_double,
               'LOGICAL1'  : c_bool, # TODO logical of other lengths
               'LOGICAL4'  : c_bool, # TODO logical of other lengths
               'LOGICAL8'  : c_bool, # TODO logical of other lengths
              }
typestrs = { c_double  : 'f8',
             c_float   : 'f4',
             c_long    : 'i8',
             c_int     : 'i4',
             c_bool    : 'b4',
             c_char_p  : 'S',
             c_complex4: 'c8',
             c_complex8: 'c16',
           }
class _DL_DT_Scalar:
    __slots__ = ('_parent', '_meta')
    def __init__(self, parent, meta):
        self._parent = parent
        self._meta   = meta
    def instantiate(self):
        addr        = self._parent.__address__ + self._meta['_offset']
        typename    = self._meta['_typename']
        dict_dt     = self._parent._derived_types[typename]
        caller_base = (self._parent._caller + '%' + self._meta['_name'] if self._parent._caller else self._meta['_name'])
        inst = DL_DT(addr,
                     dict_dt,
                     self._parent._derived_types,
                     caller=caller_base,
                     caller_type=typename,
                     return_ctype=self._parent._return_ctype,
                     debug=self._parent._debug)
        setattr(inst, '_derived_type', typename)
        return inst
class _DL_DT_Pointer:
    __slots__ = ('_parent', '_meta')
    def __init__(self, parent, meta):
        self._parent = parent
        self._meta   = meta
    def instantiate(self):
        offset   = self._meta['_offset']
        typename = self._meta['_typename']
        dict_dt  = self._parent._derived_types[typename]
        addr = c_ulong.from_address(self._parent.__address__ + offset).value
        if addr == 0:
            return
        if self._parent.lib_dl_py2f.associated(c_ulong(addr)):
            return
        caller      = self._parent._caller
        name_member = self._meta['_name']
        caller_base = (caller + '%' + name_member) if caller else name_member
        is_linked_list = False
        if caller:
            base = caller.split('%')[-1].split('(')[0]
            is_linked_list = (base == name_member)
            if is_linked_list and self._parent._debug:
                print(f' >>> NB: {_clr._bE}{caller_base}{_clr.CLR_} is a {_clr._bW}scalar linked list!{_clr.CLR_}')
        if self._parent.lib_dl_py2f.associated(c_ulong(addr)):
            if self._parent._debug:
                print(f' >>> WARNING: {_clr._R}{caller_base}{_clr.CLR_} at {_clr._uW}{hex(addr)}{_clr.CLR_} is likely to be a wild pointer!')
            return
        inst = DL_DT(addr,
                    dict_dt,
                    self._parent._derived_types,
                    caller=caller_base,
                    caller_type=typename,
                    return_ctype=self._parent._return_ctype,
                    debug=self._parent._debug)
        setattr(inst, '_derived_type', typename)
        if getattr(inst, '_return_immediately', False):
            if is_linked_list:
                self._parent._return_immediately = True
            else:
                if self._parent._debug:
                    print(f' >>> WARNING: {_clr._R}{caller_base}{_clr.CLR_} is likely to be an unassociated pointer to a linked list! (type: {_clr._Y}{typename}{_clr.CLR_})')
            return inst
        return inst
class _DL_DT_Array:
    __slots__ = ('_parent', '_meta', '_shape', '_typename', '_array')
    def __init__(self, parent, meta):
        self._parent   = parent
        self._meta     = meta
        self._shape    = tuple(meta['_shape'][::-1])
        self._typename = meta['_typename']
        self._array = None
    def __getitem__(self, ind):
        if not isinstance(ind, tuple):
            ind = (ind,)
        dims = self._shape
        if len(ind) != len(dims)    \
        or any(isinstance(x, slice) \
               or x is Ellipsis     \
               or x is None         \
               for x in ind):
            return _DL_DT_ArrayView(self, ind)
        norm = []
        for i, dim in zip(ind, dims):
            if isinstance(i, int):
                if i < 0:
                    i = dim + i
                if i < 0 or i >= dim:
                    raise IndexError(f'Index {i} out of bounds for dimension size {dim}')
            norm.append(i)
        ind = tuple(norm)
        return self.__getitem_int__(ind)
    def __getitem_int__(self, ind):
        if not isinstance(ind, tuple):
            ind = (ind,)
        dims = self._shape
        if len(ind) != len(dims):
            raise IndexError(f'Index {ind} has {len(ind)} dims, but array has {len(dims)} dims')
        norm = []
        for i, dim in zip(ind, dims):
            if isinstance(i, int):
                if i < 0:
                    i = dim + i
                if i < 0 or i >= dim:
                    raise IndexError(f'Index {i} out of bounds for dimension size {dim}')
            norm.append(i)
        ind = tuple(norm)
        flat = ind[0]
        for i, d in zip(ind[1:], dims[1:]):
            flat = flat*d + i
        offset     = self._meta['_offset']
        size_chunk = self._meta['_size_chunk']
        if self._meta['_kind'] == 'array_deferred':
            base = c_ulong.from_address(self._parent.__address__ + offset).value
            if base == 0:
                raise RuntimeError(f' >>> DL_F2PY ERROR: Deferred array "{self._meta["_name"]}" is not allocated (NULL pointer).')
        else:
            base = self._parent.__address__ + offset
        addr = base + flat*size_chunk
        dict_dt = self._parent._derived_types[self._typename]
        base_name = self._meta['_name']
        if self._parent._caller:
            caller_base = self._parent._caller + '%' + base_name
        else:
            caller_base = base_name
        if self._parent._debug:
            caller = caller_base + '(' + ','.join(str(i+1) for i in ind[::-1]) + ')'
        else:
            caller = caller_base
        inst = DL_DT(addr,
                    dict_dt,
                    self._parent._derived_types,
                    caller=caller,
                    caller_type=self._typename,
                    return_ctype=self._parent._return_ctype,
                    debug=self._parent._debug)
        setattr(inst, '_derived_type', self._typename)
        return inst
    def __iter__(self):
        for i in range(len(self)):
            yield self[i]
    def __len__(self):
        return self._shape[0] if self._shape[0] else 0
    @property
    def __array__(self):
        if self._array is not None:
            return self._array
        abuff = empty(self._shape, dtype=object)
        for ind in ndindex(*self._shape):
            abuff[ind] = self[ind]
        self._array = abuff
        return abuff
    @property
    def __shape__(self):
        return self._shape
    @property
    def size(self):
        n = 1
        for d in self._shape:
            n *= d
        return n
    @property
    def shape(self):
        return self._shape
    @property
    def ndim(self):
        return len(self._shape)
    @property
    def dtype(self):
        return dtype(object)
    @property
    def T(self):
        return self.__array__.T
class _DL_DT_ArrayView:
    __slots__ = ('_parent', '_meta', '_slices', '_shape', '_is_scalar')
    def __init__(self, parent, slices):
        self._parent = parent
        self._meta   = parent._meta
        self._slices = self._normalise_slices(slices, parent._shape)
        shape = []
        shape_parent = parent._shape
        i_parent = 0
        for slc in self._slices:
            if slc is None:
                shape.append(1)
            elif isinstance(slc, slice):
                dim = shape_parent[i_parent]
                start, stop, step = slc.indices(dim)
                if step == 0:
                    raise ValueError("Slice step cannot be zero")
                length = max(0, ceil((stop-start)/step))
                shape.append(length)
                i_parent += 1
            elif isinstance(slc, int):
                i_parent += 1
            else:
                raise TypeError(f"Unsupported index type {type(slc)}")
        self._shape = tuple(shape)
        if len(self._shape) == 0:
            self._is_scalar = True
        elif len(self._shape) == 1 and self._shape[0] == 1:
            self._is_scalar = True
        else:
            self._is_scalar = False
    def _normalise_slices(self, slices, shape):
        ndims_parent = len(shape)
        if Ellipsis in slices:
            ind   = slices.index(Ellipsis)
            left  = slices[:ind]
            right = slices[ind+1:]
            def _consumes_dim(s):
                return s is not None and s is not Ellipsis
            used = sum(1 for s in left+right if _consumes_dim(s))
            missing = ndims_parent - used
            slices = left + (slice(None),)*missing + right
        if len(slices) < len(shape):
            slices = slices + (slice(None),) * (len(shape)-len(slices))
        norm = []
        i_parent = 0
        for slc in slices:
            if slc is None:
                norm.append(None)
                continue
            if i_parent >= ndims_parent:
                raise IndexError("Too many indices for array")
            dim = shape[i_parent]
            if isinstance(slc, int):
                if slc < 0:
                    slc = dim + slc
                if slc < 0 or slc >= dim:
                    raise IndexError(f'Index {slc} out of bounds for dimension size {dim}')
                norm.append(slc)
                i_parent += 1
            elif isinstance(slc, slice):
                norm.append(slc)
                i_parent += 1
            else:
                raise TypeError(f"Unsupported index type {type(slc)}")
        while i_parent < ndims_parent:
            norm.append(slice(None))
            i_parent += 1
        return tuple(norm)
    def __getattr__(self, attr):
        if self._is_scalar:
            return self.__getitem_int__(0).__getattribute__(attr)
        return super().__getattribute__(attr)
    def __setattr__(self, attr, value):
        if attr in [ '_parent', '_meta', '_slices', '_shape' ,'_is_scalar' ]:
            return object.__setattr__(self, attr, value)
        if getattr(self, '_is_scalar', False):
            return self.__getitem_int__(0).__setattr__(attr, value)
        return object.__setattr__(self, attr, value)
    def __getitem__(self, ind):
        if isinstance(ind, tuple):
            norm = []
            for i, dim in zip(ind, self._shape):
                if isinstance(i, int) and i < 0:
                    i = dim + i
                if isinstance(i, int) and (i < 0 or i >= dim):
                    raise IndexError(f'Index {i} out of bounds for dimension size {dim}')
                norm.append(i)
            ind = tuple(norm)
            return _DL_DT_ArrayView(self, ind)
        if isinstance(ind, slice):
            return _DL_DT_ArrayView(self, (ind,))
        if isinstance(ind, int):
            if ind < 0:
                ind = self._shape[0] + ind
            if ind < 0 or ind >= self._shape[0]:
                raise IndexError(f'Index {ind} out of bounds for dimension size {self._shape[0]}')
        return self.__getitem_int__(ind)
    def __getitem_int__(self, ind):
        if isinstance(ind, int):
            if ind < 0:
                ind = self._shape[0] + ind
            if ind < 0 or ind >= self._shape[0]:
                raise IndexError(f'Index {ind} out of bounds for dimension size {self._shape[0]}')
        parent = self._parent
        shape_parent = parent._shape
        i_parent = 0
        indices = []
        used = False
        for slc in self._slices:
            if slc is None:
                continue
            dim = shape_parent[i_parent]
            if isinstance(slc, slice):
                start, stop, step = slc.indices(dim)
                if not used:
                    i = start + ind*step
                    used = True
                else:
                    i = start
                indices.append(i)
                i_parent += 1
            elif isinstance(slc, int):
                indices.append(slc)
                i_parent += 1
            else:
                raise TypeError(f"Unsupported slice type {type(slc)}")
        return parent.__getitem_int__(tuple(indices))
    def __iter__(self):
        for i in range(len(self)):
            yield self[i]
    def __len__(self):
        return self._shape[0] if self._shape[0] else 0
    @property
    def __array__(self):
        abuff = empty(self._shape, dtype=object)
        for ind in ndindex(*self._shape):
            abuff[ind] = self[ind]
        return abuff
    @property
    def __shape__(self):
        return self._shape
    @property
    def size(self):
        n = 1
        for d in self._shape:
            n *= d
        return n
    @property
    def shape(self):
        return self._shape
    @property
    def ndim(self):
        return len(self._shape)
    @property
    def dtype(self):
        return dtype(object)
    @property
    def T(self):
        return self.__array__.T
class DL_DT(_SimpleCData):
    _type_ = 'O'
    class _wrapper():
        __slots__ = ('__array_interface__',)
    def __new__(cls,
                address      :int,
                dict_dt      :dict,
                derived_types:dict,
                caller       :str  = '',
                caller_type  :str  = '',
                return_ctype :bool = False,
                debug        :bool = False):
        key = (address, id(dict_dt))
        obj = _DL_CACHE.get(key)
        if obj is not None:
            return obj
        inst = _SimpleCData.__new__(cls)
        _DL_CACHE[key] = inst
        return inst
    def __repr__(self):
        return f'<dl_f2py.DL_DT object at {self.__address__}>'
    @property
    def lib_dl_py2f(self):
        try:
            return self._lib_dl_py2f
        except:
            self._lib_dl_py2f = utils.modutils.getLinkedLib('dl_py2f', files(__package__), is_linked=True) \
                             or utils.modutils.getLinkedLib('dl_py2f', libpath.libpath, is_linked=True)
            return self._lib_dl_py2f
    def __init__(self,
                 address      :int,
                 dict_dt      :dict,
                 derived_types:dict,
                 caller       :str  = '',
                 caller_type  :str  = '',
                 return_ctype :bool = False,
                 debug        :bool = False):
        if type(dict_dt) is not dict or not dict_dt.get('_is_derived', False):
            raise RuntimeError(' >>> DL_F2PY ERROR: Argument dict_dt must be a dict containing information of data structure of a Fortran derived type', flush=True)
        self._dict_dt       = dict_dt
        self._derived_types = derived_types
        self._caller        = caller
        self._return_ctype  = return_ctype
        self._debug         = debug
        self._children      = {} 
        setattr(self, '__address__', address)
        icount = 0
        for k, v in dict_dt.items():
            if not isinstance(v, dict) or '_offset' not in v:
                continue
            icount += 1
            if '_ndims' in v:
                if v.get('_is_derived', False):
                    if v.get('_is_deferred', False):
                        offset = v['_offset']
                        ndims = v['_ndims']
                        address = c_ulong.from_address(self.__address__+offset).value
                        __address__ = self.__address__ + offset
                        lbuff = [ c_long.from_address(__address__+i*8).value for i in range(5+ndims*3) ]
                        lbounds = lbuff[6::3]
                        ubounds = lbuff[7::3]
                        shape = tuple(ubounds[x]-lbounds[x] for x in range(ndims))
                        if any([x <= 0 for x in shape]) or not lbuff[0]:
                            setattr(self, k, c_void_p(None))
                            continue
                        shape = tuple(x+1 for x in shape)
                        _is_allocated = True
                        if lbuff[2] != v['_size_chunk']:
                            _is_allocated = False
                            if debug:
                                print(f'{_clr._bR}\n >>> WARNING: = In {_clr.CLR_}{_clr._bY}{caller}{_clr.CLR_}{_clr._bR}, there may be something wrong with derived-type array {_clr._bY}{k}{_clr.CLR_}{_clr._bR}!{_clr.CLR_}')
                                print(f'{_clr._bR}              Shape: {shape}{_clr.CLR_}')
                            setattr(self, k, empty((), dtype=typestrs.get(v['_type'], 'i8')))
                            continue
                        if not _is_allocated:
                            shape = ()
                        dims = [ range(x) for x in shape[::-1] ]
                        indices = list(product(*dims))
                        indices.reverse()
                        is_linked_list = caller.split('%')[-1].split('(')[0] == k
                        is_allocated = True

                        if is_linked_list:
                            if debug:
                                print(f' >>> NB: {_clr._bE}{caller}%{k}{_clr.CLR_} is an {_clr._bW}array of linked list of a deferred shape{_clr.CLR_}! (type: {_clr._W_i}{v["_type"]}{_clr.CLR_})!')
                            for i, _ in enumerate(indices):
                                addr_base = address + (len(indices)-i-1)*v['_size_chunk']
                                llbuff = [ c_long.from_address(addr_base).value for i in range(5+ndims*3) ]
                                if llbuff[2] != v['_size_chunk']:
                                    is_allocated = False
                                    break
                                for j, ind in enumerate(indices):
                                    addr = addr_base + (len(indices)-j-1)*v['_size_chunk']
                                    llbuff = [ c_long.from_address(addr).value for i in range(5+ndims*3) ]
                                    if llbuff[2] != v['_size_chunk']:
                                        is_allocated = False
                                        break
                                if not is_allocated:
                                    break
                        if not is_allocated:
                              if debug:
                                  print(f' >>> NB: {_clr._bW}{caller+"%"+k}{_clr.CLR_} is {_clr._R_i}NOT ALLOCATED!{_clr.CLR_}')
                              value = None
                              continue
                        self._children[k] = { '_kind'      : 'array_deferred',
                                              '_name'      :  k,
                                              '_meta'      :  v,
                                              '_typename'  :  v['_type'],
                                              '_offset'    :  v['_offset'],
                                              '_shape'     :  shape,
                                              '_size_chunk':  v.get('_size_chunk'),
                                              '_dim'       :  v.get('_dim'),
                                              '_ndims'     :  v.get('_ndims', 0),
                                              '_is_pointer':  v.get('_is_pointer', False)}
                        continue
                    else:
                        shape = tuple(v['_dim']) 
                        self._children[k] = { '_kind'      : 'array_fixed',
                                              '_name'      :  k,
                                              '_meta'      :  v,
                                              '_typename'  :  v['_type'],
                                              '_offset'    :  v['_offset'],
                                              '_shape'     :  shape,
                                              '_size_chunk':  v.get('_size_chunk'),
                                              '_ndims'     :  v.get('_ndims', 0),
                                              '_is_pointer':  v.get('_is_pointer', False)}
                        continue
                else:
                    if v.get('_is_deferred', False):
                        offset = v['_offset']
                        address = c_ulong.from_address(self.__address__+offset).value
                        ndims = v['_ndims']
                        __address__ = self.__address__ + offset
                        lbuff = [ c_long.from_address(__address__+i*8).value for i in range(5+ndims*3) ]
                        lbounds = lbuff[6::3]
                        ubounds = lbuff[7::3]
                        shape = tuple(ubounds[x]-lbounds[x]+1 for x in range(ndims))
                        if any([x <= 0 for x in shape]) or not lbuff[0]:
                            setattr(self, k, c_void_p(None))
                            continue
                        if lbuff[2] != sizeof(v['_type']):
                            _is_allocated = False
                            if debug:
                                print(f'{_clr._bR}\n >>> WARNING: In {_clr.CLR_}{_clr._bY}{caller}{_clr.CLR_}{_clr._bR}, there may be something wrong with {v["_type"].__name__} array {_clr.CLR_}{_clr._bY}{k}!{_clr.CLR_}')
                                print(f'              Size is {_clr._bR}{lbuff[2]}{_clr.CLR_} which is supposed to be {_clr._bG}{sizeof(v["_type"])}{_clr.CLR_}')
                            setattr(self, k, empty((), dtype=typestrs.get(v['_type'], 'i8')))
                            continue
                        size = lbuff[2]
                        if v['_type'] in typestrs and size:
                            abuff = self._wrapper()
                            typestr = typestrs[v['_type']]
                            abuff.__array_interface__ = {'shape':shape[::-1], 'data':(address, False), 'typestr': f'<{typestr}'}
                            try:
                                aview = array(abuff, copy=False)
                                setattr(self, k, aview)
                            except:
                                continue
                    else:
                        if v['_type'] in typestrs:
                            offset = v['_offset']
                            address = self.__address__ + v['_offset']
                            abuff = self._wrapper()
                            typestr = typestrs[v['_type']]
                            abuff.__array_interface__ = {'shape':v['_dim'][::-1], 'data':(address, False), 'typestr': f'<{typestr}'}
                            try:
                                aview = array(abuff, copy=False)
                                setattr(self, k, aview)
                            except:
                                continue
            else:
                if v['_type'] in derived_types:
                    if v.get('_is_pointer', False):
                        is_linked_list = caller.split('%')[-1].split('(')[0] == k
                        if is_linked_list and debug:
                            print(f' >>> NB: {_clr._bE}{caller}%{k}{_clr.CLR_} is a {_clr._bW}scalar linked list!{_clr.CLR_}')
                        address = c_ulong.from_address(self.__address__+v['_offset']).value
                        if address:
                            if self.lib_dl_py2f.associated(c_ulong(address)):
                                if debug:
                                    print(f' >>> WARNING: {_clr._R}{caller}%{_clr.CLR_}{_clr._bR}{k}{_clr.CLR_} at ({_clr._uW}{hex(address)}{_clr.CLR_}) might be a wild pointer!')
                                value = None
                                setattr(self, k, value)
                                continue
                            else:
                                try:
                                    self._children[k] = { '_kind'      : 'pointer',
                                                          '_name'      :  k,
                                                          '_meta'      :  v,
                                                          '_typename'  :  v['_type'],
                                                          '_offset'    :  v['_offset'],
                                                          '_size_chunk':  v.get('_size_chunk'),
                                                          '_dim'       :  v.get('_dim'),
                                                          '_ndims'     :  v.get('_ndims', 0),
                                                          '_is_pointer':  v.get('_is_pointer', False)}
                                    continue
                                except RecursionError:
                                    value = None
                                    self._return_immediately = True
                                    return
                        else:
                            value = None
                            setattr(self, k, value)
                    else:
                        self._children[k] = { '_kind'      : 'scalar',
                                              '_name'      :  k,
                                              '_meta'      :  v,
                                              '_typename'  :  v['_type'],
                                              '_offset'    :  v['_offset'],
                                              '_size_chunk':  v.get('_size_chunk'),
                                              '_dim'       :  v.get('_dim'),
                                              '_ndims'     :  v.get('_ndims', 0),
                                              '_is_pointer':  v.get('_is_pointer', False)}
                        continue
                    setattr(self, k, value)
                else:
                    if v.get('_is_char', False):
                        if v.get('_is_deferred', False):
                            pass
                        else:
                            setattr(self, f'_DL_DT_CA_{k}', self.__address__+v['_offset'])
                            setattr(self, f'_DL_DT_CL_{k}', v.get('_length', 1))
                            sbuff_getter = f'return self._getStr(self._DL_DT_CA_{k}, self._DL_DT_CL_{k}, debug={debug})'
                            sbuff_setter = f'self._setStr(self._DL_DT_CA_{k}, self._DL_DT_CL_{k}, value, debug={debug}); return 0'
                            setter_args = f'{self.__address__+v["_offset"]}, {v.get("_length", 1)}'
                    else:
                        if v.get('_is_pointer', False):
                            sbuff_getter = f'return self._getPtr(self._DL_DT_P_{k}, {v["_type"].__name__}, name="{k}", return_ctype={return_ctype}, debug={debug})'
                            setattr(self, f'_DL_DT_P_{k}', self.__address__+v['_offset'])
                            sbuff_setter = f'self._setPtr(self._DL_DT_P_{k}, {v["_type"].__name__}, value, name="{k}", debug={debug}); return 0'
                            setter_args = f'self._DL_DT_{k}'
                        else:
                            sbuff_getter = ''
                            sbuff_setter = ''
                            if v['_type'] == '_tb_procedure':
                                continue
                            addr = self.__address__ + v['_offset']
                            value = v['_type'].from_address(self.__address__+v['_offset'])
                            setter_args = f'self._DL_DT_{k}'
                            setattr(self, '_DL_DT_'+k, value)
                    if return_ctype:
                        sbuff = f'''
def _fget(self):
    {sbuff_getter}
    return self._DL_DT_{k}
'''
                    else:
                        sbuff = f'''
def _fget(self):
    {sbuff_getter}
    return self._DL_DT_{k}.value
'''
                    exec(sbuff)
                    sbuff = f'''
def _fset(self, value):
    {sbuff_setter}
    setters.get(type(self._DL_DT_{k}), DL_DL.setInt)({setter_args}, value)
'''
                    exec(sbuff)
                    setattr(self, '_get_DL_DT_'+k, locals()['_fget'])
                    setattr(self, '_set_DL_DT_'+k, locals()['_fset'])
                    setattr(self.__class__, k, property(fget=getattr(self, '_get_DL_DT_'+k),
                                                        fset=getattr(self, '_set_DL_DT_'+k)))
    def printBytes(self, dict_dt={}, name='', addr=0, increment=4, size_extra=0):
        if dict_dt:
            if dict_dt.get('_is_char', False):
                increment = 1
            print()
            if name:
                print(f' Name: {name}')
            print(' Type:', dict_dt['_type'])
            addr = self.__address__ + dict_dt['_offset']
            print(' Addr:', addr)
            print(' Incr:', increment, 'Bytes')
            if 'ndims' in dict_dt:
                ndims = dict_dt['_ndims']
                if dict_dt.get('_is_deferred', False):
                    size = 40+ndims*24
            else:
                if dict_dt.get('_is_char', False):
                    size = dict_dt.get('_length', 1)
                else:
                    if dict_dt.get('_is_pointer', False):
                        size = 8
                    else:
                        size = 0
        else:
            if not addr:
                print('\n >>> ERROR: `addr` must be defined when dict is not given!')
                return
            size = 0
        for i in range(0, size+size_extra, increment):
            print(f'\n No. {int(i/4)}: Offset {i} Bytes', flush=True)
            ibuff  = c_int   .from_address(addr+i)
            lbuff  = c_long  .from_address(addr+i)
            ulbuff = c_ulong .from_address(addr+i)
            cbuff  = c_char  .from_address(addr+i)
            bbuff  = c_ubyte .from_address(addr+i)
            fbuff  = c_float .from_address(addr+i)
            dbuff  = c_double.from_address(addr+i)
            vbuff  = c_void_p.from_address(addr+i)
            wbuff  = c_wchar .from_address(addr+i)
            print(' '+'-'*70, flush=True)
            print('    address  =', addr+i)
            print('    c_int    =',  ibuff.value)
            print('    c_long   =',  lbuff.value)
            print('    c_ulong  =', ulbuff.value)
            print('    c_char   =',  cbuff.value)
            print('    c_ubyte  =',  bbuff.value)
            print('    c_float  =',  fbuff.value)
            print('    c_double =',  dbuff.value)
            print('    c_void_p =',  vbuff.value, flush=True)
            print(' '+'-'*70, flush=True)
    def _getStr(self, addr, length, debug=False):
        sbuff = ''
        for i in range(0, length, 1):
            sbuff += c_char.from_address(addr+i).value.decode()
        return sbuff
    def _setStr(self, addr, length, value, debug=False):
        DL_DL.setChar(addr, length, value)
    def _getPtr(self, addr, ctype, name='', return_ctype=False, debug=False):
        addr_target = c_ulong.from_address(addr).value
        if debug:
            self.printBytes(name=name, addr=addr, size_extra=8)
        if addr_target:
            value = ctype.from_address(addr_target)
            setattr(value, '_is_associated', True)
            if return_ctype:
                return value
            else:
                return value.value
        else:
            value = c_void_p(None)
            setattr(value, '_is_associated', False)
            return value
    def _setPtr(self, addr, ctype, value, name='', debug=False):
        attr = self._getPtr(addr, ctype, debug=debug)
        if attr:
            setters.get(type(attr), DL_DL.setInt)(attr, value)
        else:
            print(f' >>> DL_F2PY ERROR: Pointer "{name}" in {self} is not allocated/associated')
    def setValue(self, name, value):
        attr = getattr(self, name.strip().lower())
        setters.get(type(attr), DL_DL.setInt)(attr, value)
    def __iter__(self):
        return self
    def __getitem__(self, index):
        if hasattr(self, '__array__'):
            return self.__array__[index]
        else:
            raise TypeError(f'{self.__repr__()} is not subscriptable')
    def __getattr__(self, name):
        meta = self._children.get(name)
        if meta is None:
            raise AttributeError(name)
        kind = meta['_kind']
        if kind == 'scalar':
            inst = _DL_DT_Scalar(self, meta)
            obj = inst.instantiate()
            setattr(self, name, obj)
            return obj
        if kind == 'pointer':
            inst = _DL_DT_Pointer(self, meta)
            obj = inst.instantiate()
            setattr(self, name, obj)
            return obj
        if kind in [ 'array_fixed', 'array_deferred' ]:
            inst = _DL_DT_Array(self, meta)
            setattr(self, name, inst)
            return inst
        raise AttributeError(f'Unknown lazy kind {kind}')
    def __setitem__(self, index, value, **args):
        print(f' >>> DL_F2PY ERROR: Elements in array {self}')
        print( '                    cannot be replaced')
        print(f'                    (Fortran derived type: "{self._derived_type}", shape: {self.__shape__})')
    def __len__(self):
        try:
            return self.__array__.size
        except:
            raise AttributeError(f'{self.__repr__()} is scalar and has no length')
    @property
    def __shape__(self):
        if hasattr(self, '__array__'):
            return self.__array__.shape
        else:
            raise AttributeError(f'{self.__repr__()} is scalar and has no shape')
    @property
    def _return_immediately(self):
        try:
            return self.__return_immediately
        except:
            return False
    @_return_immediately.setter
    def _return_immediately(self, val:bool):
        self.__return_immediately = val
class DL_DL(CDLL):
    __modules = []
    class _DL_MOD(dict):
        def __init__(self, dl_dl, *args, **kwargs):
            object.__setattr__(self, '__dl_dl__', dl_dl)
            self.update(*args)
        def __getattr__(self, attr):
            if '_name' not in self or self.get('_name') == '__modules__':
                try:
                    return self[attr]
                except KeyError:
                    raise AttributeError(attr)
            return self.__dl_dl__.getValue(attr, module=self['_name'], return_ctype=self.__dl_dl__.return_ctype, debug=False)
        def __setattr__(self, attr, val):
            self.__dl_dl__.setValue(attr, val, module=self['_name'], debug=False)
    def __new__(cls, fullpath_to_lib, mode=RTLD_GLOBAL, debug=False):
        inst = CDLL.__new__(cls)
        inst.__init__(fullpath_to_lib, mode=mode)
        inst._fullpath_to_lib = fullpath_to_lib
        inst._derived_types = {}
        inst._instances_derived_types = {}
        return inst
    def __getattr__(self, name):
        if name in [ '_moddir', '_modules', '_lib_dl_py2f' ]:
            return object.__getattribute__(self, name)
        try:
            if name.startswith('__') and name.endswith('__'):
                raise AttributeError(name)
            func = self.__getitem__(name)
            setattr(self, name, func)
            return func
        except:
            return object.__getattribute__(self, name)
    @property
    def lib_dl_py2f(self):
        try:
            return self._lib_dl_py2f
        except:
            self._lib_dl_py2f = utils.modutils.getLinkedLib('dl_py2f', files(__package__), is_linked=True) \
                             or utils.modutils.getLinkedLib('dl_py2f', libpath.libpath, is_linked=True)
            return self._lib_dl_py2f
    @property
    def moddir(self):
        try:
            return self._moddir
        except:
            print('\n >>> ERROR: A directory containing Fortran module files has not been defined!\n')
            return []
    @moddir.setter
    def moddir(self, val):
        try:
            self._moddir
        except:
            self._moddir = []
        if os.path.isdir(val.strip()):
            if not val.strip() in self._moddir:
                self._moddir += [ val.strip() ]
        else:
            print(f'\n >>> ERROR: Directory {val.strip()} does not exist!')
    @property
    def return_ctype(self):
        try:
            return self._return_ctype
        except:
            return False
    @return_ctype.setter
    def return_ctype(self, val:bool):
        self._return_ctype = val
    @property
    def symbols(self):
        self.lib_dl_py2f.getSymbols.argtypes = [ c_void_p, c_char_p ]
        self.lib_dl_py2f.getSymbols.restype = c_char_p
        bbuff = self.lib_dl_py2f.getSymbols(self.lib_dl_py2f.dl_f2py(), self._fullpath_to_lib.strip().encode())
        lbuff = bbuff.decode().split(';')
        return [ s for s in lbuff if s and ('_MOD___copy_'       not in s
                                        and '_MOD___deallocate_' not in s
                                        and '_MOD___def_init_'   not in s
                                        and '_MOD___vtab_'       not in s
                                        and '_MOD___final_'      not in s) ]
    def searchSymbol(self, *keywords):
        symbols_all = self.symbols
        return [ s for s in symbols_all if all([ k.strip().lower() in s.strip().lower() for k in keywords ]) ]
    @property
    def modulenames(self):
        symbols_all = self.symbols
        lbuff = [ s.split('_MOD_')[0].lstrip('__') for s in symbols_all if '_MOD_' in s ]
        return sorted(list(set(lbuff)))
    @property
    def _to_parse_all(self):
        try:
            return self.__to_parse_all
        except:
            return True
    @_to_parse_all.setter
    def _to_parse_all(self, val):
         self.__to_parse_all = val
    @property
    def modules(self):
        try:
            if self._to_parse_all:
                for moddir in self.moddir:
                    self.parseAllModules(moddir, caller='modules')
            return self._modules
        except Exception:
            import traceback, sys
            traceback.print_exc()
            if not hasattr(self, '_modules'):
                self._modules = self._DL_MOD(self, {'_name': '__modules__'})
            self._to_parse_all = False
            for moddir in self.moddir:
                self.parseAllModules(moddir, caller='modules')
            return self._modules
    @modules.setter
    def modules(self, val):
        try:
            overlap = set(self._modules.keys()) & (val.keys())
            if len(overlap):
                print(f'\n >>> WARNING: The following module instances will be overwritten:\n             ', ', '.join(overlap))
            self._modules.update(val)
        except:
            if not hasattr(self, '_modules'):
                self._modules = self._DL_MOD(self, {'_name': '__modules__'})
            self._modules.update(val)
    @property
    def nsymbols(self):
        return len(self.symbols)
    def getModuleSymbols(self, module):
        symbols_all = self.symbols
        return [ s for s in symbols_all if module.strip().lower() in s.lower() ]
    def getSymbolOfModule(self, symbol, module):
        if module:
            symbols_module = self.getModuleSymbols(module)
            lbuff = [ s for s in symbols_module if s.lower().rstrip('_').endswith('_'+symbol.strip().lower()) ]
            if len(lbuff) == 0:
                raise RuntimeError(f' >>> DL_PY2F ERROR: Module {module} does not have symbol "{symbol}"\n', flush=True)
            elif len(lbuff) == 1:
                return lbuff[0]
            else:
                sbuff  = f' >>> DL_PY2F ERROR: Module {module} has more than one symbols containing string "{symbol}":'
                sbuff += '\n                    ', ' '.join(lbuff)
                raise RuntimeError(sbuff)
        else:
            return symbol
    def getValue(self, symbol, ctype=c_int, shape=(), module='', return_ctype=False, debug=False):
        symbol_original = symbol
        lbuff = []
        if '%' in symbol:
            lbuff = symbol.split('%')
            symbol = lbuff[0]
        fullname_symbol = self.getSymbolOfModule(symbol, module)
        _instances_derived_types = self._instances_derived_types.get(module, {})
        derived_type = _instances_derived_types.get(symbol, None)
        ctype = selectcases.get(ctype, ctype)
        try:
            mod = getattr(self.modules, module)
        except:
            for d in self.moddir:
                mod = self.parseModule(os.path.join(d, module+'.mod'), debug=debug)
                if mod != None:
                    break
        try:
            dbuff = mod.get(symbol, {})
        except UnboundLocalError:
            dbuff = {}
        ctype = dbuff.get('_type', ctype)
        shape = dbuff.get('_dim' , shape)
        entity = ctype.in_dll(self, fullname_symbol)
        if shape != ():
            if dbuff.get('_is_deferred', False):
                if dbuff.get('_is_char', False):
                    if not dbuff.get('_is_deferred_char', False):
                        lchar = dbuff.get('_length', 1)
                addr = c_ulong.in_dll(self, fullname_symbol).value
            else:
                if dbuff.get('_is_char', False):
                    if not dbuff.get('_is_deferred_char', False):
                        lchar = dbuff.get('_length', 1)
                addr = addressof(entity)
            abuff = DL_DT._wrapper()
            typestr = typestrs[ctype]
            if ctype == c_char_p:
                typestr += str(lchar)
            abuff.__array_interface__ = {'shape':shape[::-1], 'data':(addr, False), 'typestr': f'<{typestr}'}
            aview = array(abuff, copy=False)
            return aview
        accessor = None
        accessor_key = None
        if lbuff:
            accessor_key = (module, symbol_original)
            accessor = _accessor_cache.get(accessor_key)
        if derived_type in self._derived_types.get(module, {}):
            entity = DL_DT(addressof(entity),
                           self._derived_types[module][derived_type],
                           self._derived_types[module],
                           caller=fullname_symbol,
                           caller_type=derived_type,
                           return_ctype=return_ctype,
                           debug=debug)
            if lbuff:
                if accessor is None:
                    path = []
                    for attr in lbuff[1:]:
                        attr = attr.strip()
                        if '(' in attr and ')' in attr:
                            name = attr.split('(')[0].strip()
                            index = attr.split('(')[1].split(')')[0]
                            index = eval(f'({index},)')
                            index = tuple(i-1 for i in index)[::-1]
                            path.append(('indexed', name, index))
                        else:
                            path.append(('attr', attr))
                    def _accessor(obj, _path=tuple(path)):
                        res = obj
                        for kind, *spec in _path:
                            if kind == 'attr':
                                res = getattr(res, spec[0])
                            elif kind == 'indexed':
                                name, ind = spec
                                res = getattr(res, name)[ind]
                        return res
                    _accessor_cache[accessor_key] = _accessor
                    accessor = _accessor
                entity = accessor(entity)
            return entity
        if module.strip():
            mod = getattr(self.modules, module)
            if symbol in mod:
                try:
                    ctype = mod[symbol]['_type']
                except KeyError:
                    print(f' >>> ERROR: The data type of symbol {symbol} in module {module} is not defined! Please contact {_clr.__email__}.')
                if mod[symbol].get('_is_pointer', False):
                    if return_ctype:
                        return ctype.from_address(c_ulong.in_dll(self, fullname_symbol).value)
                    else:
                        return ctype.from_address(c_ulong.in_dll(self, fullname_symbol).value).value
                else:
                    entity = ctype.in_dll(self, fullname_symbol)
            else:
                print(f' >>> ERROR: Symbol {symbol} not found in module {module}!')
                return
        if return_ctype:
            return entity
        else:
            return entity.value
    def setValue(self, symbol, value, module='', debug=False):
        if module.strip():
            mod = getattr(self.modules, module)
            if symbol in mod:
                try:
                    ctype = mod[symbol]['_type']
                    entity = self.getValue(symbol, ctype, module=module, return_ctype=True, debug=debug)
                    setters.get(type(entity), self.setInt)(entity, value)
                except KeyError:
                    print(f' >>> ERROR: The data type of symbol {symbol} in module {module} is not defined! Please contact {_clr.__email__}.')
        return
    @staticmethod
    def setInt(entity, value):
        lbuff = utils.ctypeutils.int2intarray(value)
        addr = addressof(entity)
        for i in range(4):
            memset(addr+i, lbuff[i], 1)
    @staticmethod
    def setLong(entity, value):
        lbuff = utils.ctypeutils.long2intarray(value)
        addr = addressof(entity)
        for i in range(8):
            memset(addr+i, lbuff[i], 1)
    @staticmethod
    def setChar(addr, length, value):
        try:
            value = value.encode()
        except:
            pass
        for i in range(length):
            if i < len(value):
                memset(addr+i, value[i], 1)
            else:
                memset(addr+i, 32, 1)
    @staticmethod
    def setBool(entity, value):
        lbuff = utils.ctypeutils.bool2intarray(value)
        addr = addressof(entity)
        memset(addr, lbuff[0], 1)
        for i in range(1,4):
            memset(addr+i, 0, 1)
    @staticmethod
    def setFloat(entity, value):
        lbuff = utils.ctypeutils.float2intarray(value)
        addr = addressof(entity)
        for i in range(4):
            memset(addr+i, lbuff[i], 1)
    @staticmethod
    def setDouble(entity, value):
        lbuff = utils.ctypeutils.double2intarray(value)
        addr = addressof(entity)
        for i in range(8):
            memset(addr+i, lbuff[i], 1)
    def parseAllModules(self, moddir, recursive=True, padding=True, overwrite=False, caller='', debug=False):
        from os import path, sep, walk
        if not hasattr(self, '_modules'):
            self._modules = self._DL_MOD(self, {'_name': '__modules__'})
        mods = self._modules
        items = [(m, d) for (m, d) in mods.items() if isinstance(d, dict)]
        self._to_parse_all = False
        if recursive:
            for dir_full, subdirs, filenames in walk(moddir):
                for filename in filenames:
                    if debug:
                        print(f' >>> Parsing {_clr._uW_}{path.join(dir_full, filename)}{_clr.CLR_}')
                    filepath = path.join(dir_full, filename).strip()
                    basename = utils.fileutils.getBaseName(path.basename(filepath))
                    to_parse = True
                    for m, d in items:
                        if not isinstance(d, dict):
                            continue
                        f = d.get('_filename', '')
                        if filename == f and basename == m and not overwrite:
                            to_parse = False
                            continue
                    if to_parse:
                        mod = self.parseModule(filepath, debug=debug)
                        mods.update({mod['_name']:mod})
                        object.__setattr__(mods, mod['_name'], mod)
                        self._to_parse = True
        else:
            pass
        return mods
    def parseModule(self, modpath, padding=True, debug=False):
        import gzip
        from time  import time
        from math  import prod
        from os    import path
        if not path.isfile(modpath):
            return
        fortran_internals = [ '__vtab_', '__vtype_', '__def_init_', '__copy_', '__final_', '__convert' ]
        basename = utils.fileutils.getBaseName(path.basename(modpath))
        def _printSummary(_dbuff):
            print('\n '+'*'*72)
            print(f" {_clr._bW}Summary of module{_clr.CLR_} '{_clr._bB_b}{dbuff['_name']}{_clr.CLR_}'")
            print(' '+'*'*72)
            print(f' File path: {_clr._u_W_}{path.abspath(modpath)}{_clr.CLR_}')
            print(f' {dbuff["_title"]}')
            for _k, _v in _dbuff.items():
                if _k == '_title':
                    continue
                if type(_v) is dict:
                    print(f'\n Entry "{_clr._bC}{_clr._k}{_clr.CLR_}":')
                    len_keys = max([len(_) for _ in _v.keys()])
                    fmt_keys = f'     {_clr._bY}' + '{_clr._kk:<' + f'{len_keys}' + '}' + f'{_clr.CLR_}'
                    for _kk, _vv in _v.items():
                        if _kk.startswith('_'):
                            if type(_vv) is int:
                                fmt_vv = f': {_clr._bM_b}'+'{_clr._vv}'+f'{_clr.CLR_}'
                            elif type(_vv) is bool:
                                fmt_vv = f': {_clr._G}'+'{_clr._vv}'+f'{_clr.CLR_}'
                            elif type(_vv) is str:
                                fmt_vv = f': {_clr._C}'+"'{_clr._vv}'"+f'{_clr.CLR_}'
                            elif isinstance(_SimpleCData, type(_vv)):
                                fmt_vv = f': {_clr._R}'+'{_clr._vv}'+f'{_clr.CLR_}'
                            else:
                                fmt_vv = ': {_clr._vv}'
                            print(f'{fmt_keys+fmt_vv}'.format(_kk=_kk, _vv=_vv))
                    fmt_keys += ': {_clr._vv}'
                    for _kk, _vv in _v.items():
                        if not _kk.startswith('_'):
                            if type(_vv) is dict:
                                print(f'{fmt_keys}'.format(_kk=_kk, _vv=''), end='')
                                for _iii, (_kkk,_vvv) in enumerate(_vv.items()):
                                    if _iii:
                                        print(' '*(len_keys+7), end='')
                                    if type(_vvv) is int:
                                        if _kkk == '_offset':
                                            print(f'''{_clr._bC}{"'"+_kkk+"'":17s}{_clr.CLR_}: {_clr._bY_r}{_clr._vvv}{_clr.CLR_}''')
                                        else:
                                            print(f'''{_clr._C}{"'"+_kkk+"'":17s}{_clr.CLR_}: {_clr._M}{_clr._vvv}{_clr.CLR_}''')
                                    elif type(_vvv) is bool:
                                        print(f'''{_clr._C}{"'"+_kkk+"'":17s}{_clr.CLR_}: {_clr._G}{_clr._vvv}{_clr.CLR_}''')
                                    elif type(_vvv) is str:
                                        print(f'''{_clr._C}{"'"+_kkk+"'":17s}{_clr.CLR_}: {_clr._bY}{_clr._vvv}{_clr.CLR_}''')
                                    elif type(_vvv) is tuple:
                                        print(f'''{_clr._C}{"'"+_kkk+"'":17s}{_clr.CLR_}: {_clr._vvv}''')
                                    elif isinstance(_SimpleCData, type(_vvv)):
                                        print(f'''{_clr._C}{"'"+_kkk+"'":17s}{_clr.CLR_}: {_clr._bGbE}{_clr._vvv}{_clr.CLR_}''')
                                    else:
                                        print(f'''{_clr._C}{"'"+_kkk+"'":17s}{_clr.CLR_}: {_clr._R}{_clr._vvv}{_clr.CLR_}''')
                            else:
                                print(f'{fmt_keys}'.format(_kk=_kk, _vv=_vv))
                else:
                    print(f'\n Entry "{_clr._k}": {_clr._v}')
            print(f'\n >>> Total time used: {time()-t0} s\n')
        def _recalcOffset(_dict_member, _sizes, _offset, _padding):
            if not _dict_member.get('_is_pending', False):
                return _offset, False
            _ndims = int(_dict_member.get('_ndims', 0) or 0)
            if _ndims:
                if _dict_member.get('_is_deferred', False):
                    if '_stride' not in _dict_member or '_size' not in _dict_member:
                        return _offset, False
                    _stride = _dict_member['_stride']
                    _size   = _dict_member['_size']
                    _residue = (sum(_sizes)%_stride) if _stride else 0
                    _pad = 0
                    if _dict_member.get('_is_char', False):
                        if _padding and _residue:
                            _pad = abs(_stride - _residue)
                    else:
                        if _residue and _size > _residue:
                            _pad = abs(_stride - _residue)
                    if _pad and _sizes:
                        _sizes[-1] += _pad
                        _dict_member.update({'_padded': _pad, '_is_padded': True})
                    _offset += (_sizes[-1] if _sizes else 0)
                    _dict_member.update({'_offset': _offset})
                    _sizes += [40 + 24*_ndims]
                    _dict_member.pop('_is_pending', None)
                    return _offset, True
                if not _dict_member.get('_dim', None) or '_size' not in _dict_member:
                    return _offset, False
                _offset += (_sizes[-1] if _sizes else 0)
                _dict_member.update({'_offset': _offset})
                _sizes += [_dict_member['_size'] * prod(_dict_member['_dim'])]
                _dict_member.pop('_is_pending', None)
                return _offset, True
            if '_size' not in _dict_member:
                return _offset, False
            _offset += (_sizes[-1] if _sizes else 0)
            _sizes += [_dict_member['_size']]
            _dict_member.update({'_offset': _offset})
            _dict_member.pop('_is_pending', None)
            return _offset, True
        dbuff = self._DL_MOD(self, {'_name':basename})
        indices2entries = {}
        _instances_derived_types = {}
        try:
            fp = gzip.open(modpath)
            fp.read()
        except:
            return
        with gzip.open(modpath) as mod, open(f'._dl_py2f_{basename}.txt', 'w') as fp:
            sbuff = ' '.join(mod.read().decode().replace('()', '').replace('\'\'', '').split())
            fp.write('DL_PY2F\nLegend to hierarchies:\n')
            fp.write('{[(⟨⟪<“”>⟫⟩)]}\n\n')
            is_done    = True
            depth  = 0
            offset = 0
            sbuff0 = ''
            sbuff1 = ''
            sbuff2 = ''
            sbuff3 = ''
            sbuff4 = ''
            sbuff5 = ''
            sbuff6 = ''
            sbuff7 = ''
            sizes  = []
            types2ids   = {}
            ids2types   = {}
            types2sizes = {'_tb_procedure':0}
            for ic, c in enumerate(sbuff):
                if c == '(':
                    depth += 1
                elif c == ')':
                    depth -= 1
                if c == '(':
                    if depth == 1:
                        sbuff1 = ''
                        if sbuff0:
                            dbuff.update({'_title':sbuff0})
                        fp.write('\n{')
                    elif depth == 2:
                        sbuff2 = ''
                        self.__entry = ''
                        try:
                            modname = sbuff1.replace('(','').replace(')','').strip().split()[1].strip("'")
                        except:
                            modname = ''
                        try:
                            lsbuff1 = sbuff1.replace('(','').replace(')','').strip().split()
                            index = int(lsbuff1[0])
                            entry = lsbuff1[1].strip("'")
                            if len(lsbuff1) == 4:
                                module = lsbuff1[2].strip("'")
                            elif len(lsbuff1) == 3:
                                module = ''
                        except:
                            if is_done:
                                entry = sbuff1.replace('(','').replace(')','').strip()
                                module = ''
                        is_internal = any([entry.startswith(x) for x in fortran_internals])
                        is_internal += entry.lower().strip() in [ 'c_ptr', 'c_funptr' ]
                        if is_internal:
                            fp.write('\n    [')
                            continue
                        if entry.islower():
                            if entry in dbuff and is_done:
                                entry += f'_{index}'
                        else:
                            entry = entry.lower()
                        if entry:
                            self.__entry = entry
                            if is_done:
                                if is_internal:
                                    dict_entry = {}
                                else:
                                    dbuff.update({entry.lower():{'_index':index}})
                                    dict_entry = dbuff[entry.lower()]
                                if module:
                                    if module not in self.__modules:
                                        self.__modules += [ module ]
                                    dict_entry.update({'_module':module})
                            indices2entries.update({index:entry.lower()})
                            size = 0
                            stride = 0
                            sizes = []
                            procedures = {}
                            fp.write('\n    [')
                        else:
                            fp.write('\n    [')
                    elif depth == 3:
                        sbuff3 = ''
                        if is_internal:
                            fp.write('\n        (')
                            continue
                        fp.write('\n        (')
                    elif depth == 4:
                        sbuff4 = ''
                        lbuff3 = sbuff3.split()
#                        self.__printSbuff(sbuff3, sbuff=3, depth=4, entry='arr03_of_int', debug=debug)
                        if is_internal:
                            fp.write('\n            ⟨')
                            continue
                        if dict_entry.get('_is_const', False):
                            if sbuff3.strip() == 'ARRAY':
                                dict_entry.update({'_ndims':1})
                            if dict_entry.get('_ndims', 0):
                                try:
                                    dict_entry.update({'_ndims':int(sbuff3)})
                                    dict_entry.update({'_lbuff': []})
                                except:
                                    pass
                            if len(lbuff3) == 6:
                                if lbuff3[0] == 'CHARACTER' and lbuff3[-1] == 'CHARACTER':
                                    dict_entry.update({'_is_char':True})
                                    dict_entry.update({'_type':selectcases[lbuff3[0]+lbuff3[1]]})
                                    dict_entry.update({'_length': int(lbuff3[1].strip("'"))})
                        else:
                            if len(lbuff3) == 6:
                                if lbuff3[0] == 'CHARACTER' and lbuff3[-1] == 'CHARACTER':
                                    dict_entry.update({'_is_char':True})
                                    dict_entry.update({'_type':selectcases[lbuff3[0]+lbuff3[1]]})
                                    dict_entry.update({'_length': int(lbuff3[1].strip("'"))})
                            if dict_entry.get('_is_char', False):
                                if len(lbuff3) == 3:
                                    if lbuff3[-1] == 'EXPLICIT':
                                        dict_entry.update({'_ndims':int(lbuff3[0])})
                                        dict_entry.update({'_lbuff': []})
                            else:
                                if len(lbuff3) == 3:
                                    if lbuff3[-1] == 'EXPLICIT':
                                        dict_entry.update({'_ndims':int(lbuff3[0])})
                                        dict_entry.update({'_lbuff': []})
                        fp.write('\n            ⟨')
                    elif depth == 5:
                        sbuff5 = ''
                        lbuff4 = sbuff4.split()
#                        self.__printSbuff(sbuff4, sbuff=4, depth=5, entry='arr03_of_int', debug=debug)
                        if is_internal:
                            fp.write('\n                ⟪')
                            continue
                        try:
                            name = lbuff4[1].strip("'")
                            idx  = int(lbuff4[0])
                        except:
                            name = sbuff4
                            idx  = None
                        if sbuff4.strip():
                            if entry and not is_internal:
                                dbuff.update({entry:dict_entry})
                            if len(name.split()) == 1:
                                if name.strip() not in [ 'UNKNOWN-ACCESS', 'CONSTANT', 'c_address', 'PUBLIC', 'PRIVATE', 'PROTECTED', 'PLUS', 'MINUS' ]:
                                    name_member = name
                                    dict_entry.update({name_member:{'_index':idx}})
                                    dict_member = dict_entry[name_member]
                        try:
                            dict_member
                        except:
                            dict_entry.update({'_no_member':True})
                            dict_member = {}
                            name_member = ''
                        else:
                            dict_entry.pop('_no_member', None)
                        fp.write('\n                ⟪')
                    elif depth == 6:
                        sbuff6 = ''
                        lbuff5 = sbuff5.split()
#                        self.__printSbuff(sbuff5, sbuff=5, depth=6, entry='arr03_of_int', debug=debug)
                        if is_internal or dict_entry.get('_no_member', False) or 'dict_member' not in locals():
                            sbuff6 = ''
                            fp.write('\n                    <')
                            continue
                        if len(lbuff5) == 3:
                            if lbuff5[-1] == 'EXPLICIT':
                                dict_member.update({'_ndims':int(lbuff5[0])})
                                dict_member.update({'_lbuff': []})
                                dict_member.update({'_forced_pad':True})
                        elif len(lbuff5) == 6:
                            if lbuff5[0] == 'CHARACTER' and lbuff5[-1] == 'CHARACTER':
                                dict_member.update({'_is_char':True})
                                dict_member.update({'_type':selectcases[lbuff5[0]+lbuff5[1]]})
                                dict_member.update({'_length': int(lbuff5[1].strip("'"))})
                            else:
                                key = lbuff5[0] + lbuff5[1]
                                if key in selectcases:
                                    dt = selectcases[key]
                                    nbytes = int(lbuff5[1])
                                    if lbuff5[0] == 'COMPLEX':
                                        nbytes *= 2
                                    dict_member.update({'_type': dt})
                                    dict_member.update({'_nbytes': nbytes})
                                    dict_member.update({'_size'  : nbytes})
                                    dict_member.update({'_stride': nbytes})
                                    types2sizes.update({dt: nbytes})
                                    if dict_member.get('_is_pending', False):
                                        offset, offset_done = _recalcOffset(dict_member, sizes, offset, padding)
                        else:
                            pass
                        fp.write('\n                    <')
                    else:
                        sbuff7 = ''
                        if is_internal:
                            fp.write('\n                        “')
                            continue
                        fp.write('\n                        “')
                elif c == ')':
                    if depth == 0:
                        fp.write('\n}')
                        sbuff1 = ''
                    elif depth == 1:
                        sbuff1 = ''
                        if is_internal:
                            fp.write('\n    ]')
                            continue
                        if entry:
                            types2sizes.update({entry:sum(sizes)})
                            if 'MODULE' in sbuff3:
                                dict_entry.update({'_is_module':True})
                        else:
                            is_type    = False
                        offset = 0
                        try:
                            if is_done and entry and not is_internal:
                                dbuff.update({entry:dict_entry})
                        except:
                            pass
                        fp.write('\n    ]')
                        sbuff2 = ''
                    elif depth == 2:
                        if is_internal:
                            sbuff3 = ''
                            fp.write('\n        )')
                            continue
                        sbuff2 = ''
                        lbuff3 = sbuff3.split()
#                        self.__printSbuff(sbuff3, sbuff=3, depth=2, entry='arr03_of_int', debug=debug)
                        if 'PARAMETER' in lbuff3:
                            dict_entry.update({'_is_const':True})
                        if 'VARIABLE' in lbuff3:
                            dict_entry.update({'_is_var':True})
                            if lbuff3[-1] == 'POINTER':
                                dict_entry.update({'_is_pointer':True})
                            if lbuff3[-1] == 'TARGET':
                                dict_entry.update({'_is_target':True})
                        if dict_entry.get('_is_const', False) or dict_entry.get('_is_var', False):
                            if len(lbuff3) == 6:
                                if lbuff3[0] + lbuff3[1] in selectcases:
                                    dt = selectcases[lbuff3[0]+lbuff3[1]]
                                    dict_entry.update({'_type':dt})
                            if len(lbuff3) == 2 and lbuff3[1].startswith("'") and lbuff3[1].endswith("'"):
                                hexadecimal = lbuff3[1].strip("'")
                                value = self.hexadecimal2number(hexadecimal, dict_entry['_type'])
                                dict_entry.update({'_value':value})
                                object.__setattr__(dbuff, entry, value)
                            if dict_entry.get('_is_char', False):
                                if len(sbuff3.split("'")) == 3:
                                    dict_entry.update({'_length':int(lbuff3[1])})
                                    dict_entry.update({'_value':sbuff3.split("'")[-2][:dict_entry['_length']]})
                                    object.__setattr__(dbuff, entry, dict_entry['_value'])
                        if len(lbuff3) > 1:
                            if lbuff3[0] == 'DERIVED':
                                dict_entry.update({'_is_derived':True})
                                if not lbuff3[1].isdigit():
                                    if index not in ids2types and not is_internal:
                                        ids2types.update({index:entry})
                                        types2ids.update({entry:index})
                                        dict_entry.update({'_type_id':index})
                            if lbuff3[0]+lbuff3[1] in selectcases:
                                dt = selectcases[lbuff3[0]+lbuff3[1]]
                                dict_entry.update({'_type':dt})
                                types2sizes.update({dt:int(lbuff3[1])})
                            if 'DEFERRED' in lbuff3:
                                dict_entry.update({'_is_deferred':True})
                                dict_entry.update({'_ndims':int(lbuff3[0])})
                        if dict_entry.get('_is_const', True):
                            if dict_entry.get('_type', None) == c_bool:
                                if len(lbuff3) == 2:
                                    hexadecimal = lbuff3[-1]
                                    value = self.hexadecimal2number(hexadecimal, dict_entry['_type'])
                                    dict_entry.update({'_value':value})
                                    object.__setattr__(dbuff, entry, value)
                        fp.write('\n        )')
                        try:
                            index_derived_type = int(sbuff3.partition('DERIVED')[2].split()[0])
                            tmp = indices2entries.get(index_derived_type, index_derived_type)
                            if not entry.startswith('__vtab'):
                                _instances_derived_types.update({entry:tmp})
                        except:
                            pass
                        sbuff3 = ''
                    elif depth == 3:
                        sbuff3 = ''
                        lbuff4 = sbuff4.split()
#                        self.__printSbuff(sbuff4, sbuff=4, depth=3, entry='arr03_of_int', debug=debug)
                        if is_internal:
                            sbuff4 = ''
                            fp.write('\n            ⟩')
                            continue
                        if dict_entry.get('_ndims', 0):
                            if dict_entry.get('_is_const', False):
                                if len(lbuff4) == dict_entry['_ndims']:
                                    shape = [ int(_.strip("'")) for _ in lbuff4 ][::-1]
                                    if all(shape):
                                        dict_entry.update({'_dim':tuple(shape)})
                                        dict_entry.update({'_value': full((len(dict_entry['_lbuff']),),
                                                                          dict_entry['_lbuff'],
                                                                          dtype=dict_entry['_type']).reshape(shape, order='C')})
                                        dict_entry.pop('_lbuff')
                                        object.__setattr__(dbuff, entry, dict_entry['_value'])
                            else:
                                if len(lbuff4) == 2 and lbuff4[1].startswith("'") and lbuff4[1].endswith("'"):
                                    _token = lbuff4[-1].strip("'")
                                    if _token.lstrip('+-').isdigit():
                                        dict_entry['_lbuff'] += [ int(_token) ]
                                    else:
                                        dict_entry.setdefault('_dim_expr', []).append(_token)
                                        dict_entry.update({'_dim': None})
                                        continue
                                    if not len(dict_entry['_lbuff'])%2:
                                        dict_entry.update({'_dim':tuple(dict_entry['_lbuff'][i*2+1]-dict_entry['_lbuff'][i*2]+1 for i in range(len(dict_entry['_lbuff'][::2])))})
                                        if len(dict_entry['_dim']) == dict_entry['_ndims']:
                                            dict_entry.pop('_lbuff')
                        try:
                            name_member
                        except UnboundLocalError:
                            name_member = ''
                        try:
                            dict_member
                        except UnboundLocalError:
                            continue
                        if not name_member.strip():
                            continue
                        if dict_member.get('_ndims', 0):
                            need_dim = (not dict_member.get('_is_deferred', False))
                            dict_member.update({'_forced_pad':True})
                            if dict_member.get('_is_deferred', False):
                                if dict_member.get('_is_char', False):
                                    residue = sum(sizes)%dict_member['_stride']
                                    if padding and residue:
                                        if debug > 1:
                                            print(f' >>> {_clr._bW}Padding{_clr.CLR_} {_clr._bC}{entry}%{_clr.CLR_}{_clr._bY}{name_member}{_clr.CLR_} by {_clr._bM}{dict_member["_stride"]-residue}{_clr.CLR_}:')
                                        pad = abs(dict_member['_stride'] - residue)
                                        sizes[-1] += pad
                                        dict_member.update({'_padded':pad})
                                        dict_member.update({'_is_padded':True})
                                    try:
                                        offset += sizes[-1]
                                    except IndexError:
                                        offset += 0
                                else:
                                    residue = sum(sizes)%dict_member['_stride']
                                    if residue and dict_member['_size'] > residue:
                                        if debug > 1 and not is_internal:
                                            print(f' >>> {_clr._bW}Padding{_clr.CLR_} {_clr._bC}{entry}%{_clr.CLR_}{_clr._bY}{name_member}{_clr.CLR_} by {_clr._bM}{dict_member["_stride"]-residue}{_clr.CLR_} bytes:')
                                            print(f' >>>     Size: {_clr._bM}{dict_member["_size"]}{_clr.CLR_}, Stride: {_clr._bM}{dict_member["_stride"]}{_clr.CLR_}, Residue: {_clr._bM}{residue}{_clr.CLR_}')
                                        pad = abs(dict_member['_stride'] - residue)
                                        sizes[-1] += pad
                                        dict_member.update({'_is_padded':True})
                                        dict_member.update({'_padded':pad})
                                    try:
                                        offset += sizes[-1]
                                    except IndexError:
                                        offset += 0
                                dict_member.update({'_offset':offset})
                                sizes += [40+24*dict_member['_ndims']]
                            else:
                                if dict_member.get('_is_char', False):
                                    try:
                                        offset += sizes[-1]
                                    except IndexError:
                                        offset += 0
                                    sizes += [dict_member['_size']*prod(dict_member['_dim'])]
                                else:
                                    dict_member.update({'_forced_pad':True})
                                    if padding and dict_member['_stride'] and dict_member['_size']:
                                        residue = sum(sizes)%dict_member['_stride']
                                        if residue and dict_member['_size'] > residue:
                                            if debug > 1 and not is_internal:
                                                print(f' >>> {_clr._bW}Padding{_clr.CLR_} {_clr._bC}{entry}%{_clr.CLR_}{_clr._bY}{name_member}{_clr.CLR_} by {_clr._bM}{dict_member["_stride"]-residue}{_clr.CLR_} bytes:')
                                                print(f' >>>     Size: {_clr._bM}{dict_member["_size"]}{_clr.CLR_}, Stride: {_clr._bM}{dict_member["_stride"]}{_clr.CLR_}, Residue: {_clr._bM}{residue}{_clr.CLR_}')
                                            pad = abs(dict_member['_stride'] - residue)
                                            sizes[-1] += pad
                                            dict_member.update({'_is_padded':True})
                                            dict_member.update({'_padded':pad})
                                    try:
                                        offset += sizes[-1]
                                    except IndexError:
                                        offset += 0
                                    sizes += [dict_member['_size']*prod(dict_member['_dim'])]
                        else:
                            if dict_member.get('_is_char', False):
                                if dict_member.get('_is_deferred_char', False):
                                    residue = sum(sizes)%dict_member['_stride']
                                    if padding and residue:
                                        if debug > 1:
                                            print(f' >>> {_clr._bW}Padding{_clr.CLR_} {_clr._bC}{entry}.{_clr.CLR_}{_clr._bY}{name_member}{_clr.CLR_} by {_clr._bM}{stride-residue}{_clr.CLR_}:')
                                        pad = abs(dict_member['_stride'] - residue)
                                        sizes[-1] += pad
                                        dict_member.update({'_is_padded':True})
                                        dict_member.update({'_padded':pad})
                                    try:
                                        offset += sizes[-1]
                                    except:
                                        offset += 0
                                    sizes += [8]
                                else:
                                    try:
                                        offset += sizes[-1]
                                    except IndexError:
                                        offset += 0
                                    sizes += [dict_member['_size']]
                                dict_member.update({'_offset':offset})
                            else:
                                no_padding = False
                                if all([ s==4 for s in sizes ]):
                                    no_padding = True
                                if dict_member.get('_is_pointer', False):
                                    dict_member.update({'_stride':8})
                                    dict_member.update({'_size':8})
                                    residue = sum(sizes)%dict_member['_stride']
                                    if padding and residue and not no_padding:
                                        if debug > 1:
                                            print(f' >>> {_clr._bW}Padding{_clr.CLR_} {_clr._bY}{entry}.{_clr.CLR_}{_clr._bR}{name_member}{_clr.CLR_} by {_clr._bM}{stride-residue}{_clr.CLR_}:')
                                        pad = abs(dict_member['_stride'] - residue)
                                        sizes[-1] += pad
                                        dict_member.update({'_is_padded':True})
                                        dict_member.update({'_padded':pad})
                                else:
                                    if dict_member.get('_is_derived', False):
                                        dict_member.update({'_is_pointer':False})
                                        dict_member.update({'_forced_pad':True})
                                        dict_member.update({'_stride':8})
                                        dict_member.update({'_size':8})
                                        residue = sum(sizes)%dict_member['_stride']
                                        if padding and residue:
                                            pad = abs(dict_member['_stride'] - residue)
                                            sizes[-1] += pad
                                            dict_member.update({'_is_padded':True})
                                            dict_member.update({'_padded':pad})
                                            if debug > 1:
                                                print(f' >>> {_clr._bW}Padding{_clr.CLR_} {_clr._bY}{entry}%{_clr.CLR_}{_clr._bR}{name_member}{_clr.CLR_} by {_clr._bM}{pad}{_clr.CLR_} bytes:')
                                    else:
                                        if ('_stride' not in dict_member) or ('_size' not in dict_member):
                                            dict_member['_is_pending'] = True
                                            dict_member.setdefault('_offset', offset)
                                            dict_member['_name'] = name_member
                                            sbuff3 = ''
                                            sbuff4 = ''
                                            continue
                                        residue = sum(sizes)%dict_member['_stride']
                                        if residue and dict_member['_size'] > residue and not all([ s==4 for s in sizes ]):
                                            pad = abs(dict_member['_stride'] - residue)
                                            sizes[-1] += pad
                                            dict_member.update({'_is_padded':True})
                                            dict_member.update({'_padded':pad})
                                            if debug > 1 and not is_internal:
                                                print(f' >>> {_clr._bW}Padding{_clr.CLR_} {_clr._bC}{entry}%{_clr.CLR_}{_clr._bY}{name_member}{_clr.CLR_} by {_clr._bM}{pad}{_clr.CLR_} bytes:')
                                                print(f' >>>     Size: {_clr._bM}{size}{_clr.CLR_}, Stride: {_clr._bM}{dict_member["_stride"]}{_clr.CLR_}, Residue: {_clr._bM}{residue}{_clr.CLR_}')
                                try:
                                    offset += sizes[-1]
                                except IndexError:
                                    offset += 0
                                sizes += [dict_member['_size']]
                                if debug > 1 and not is_internal:
                                    print(f' >>> {_clr._bC}{entry}.{_clr.CLR_}{_clr._bY}{name_member}{_clr.CLR_} offset: {_clr._bM}{offset}{_clr.CLR_}')
                        dict_member.update({'_offset': offset})
                        fp.write(f'\n            ⟩')
                        name_member = ''
                    elif depth == 4:
                        if is_internal:
                            sbuff5 = ''
                            fp.write('\n                ⟫')
                            continue
                        sbuff4 = ''
                        lbuff5 = sbuff5.split()
                        if dict_entry.get('_is_char', False):
                            if len(lbuff5) == 2 and lbuff5[1].startswith("'") and lbuff5[1].endswith("'"):
                                dict_entry.update({'_length': int(lbuff5[1].strip("'"))})
                        if dict_entry.get('_no_member', False) or 'dict_member' not in locals():
                            sbuff5 = ''
                            fp.write('\n                ⟫')
                            continue
                        if not name_member.strip() and dict_member.get('_name'):
                            name_member = dict_member['_name']
                        if len(lbuff5) == 0:
                            pass
                        elif len(lbuff5) == 6:
                            if lbuff5[0]+lbuff5[1] in selectcases:
                                dt = selectcases[lbuff5[0]+lbuff5[1]]
                                nbytes = int(lbuff5[1])
                                if lbuff5[0] == 'COMPLEX':
                                    nbytes *= 2
                                dict_member.update({'_type':dt})
                                dict_member.update({'_nbytes': nbytes})
                                dict_member.update({'_size'  : nbytes})
                                dict_member.update({'_stride': nbytes})
                                types2sizes.update({dt       : nbytes})
                            else:
                                if lbuff5[0] == 'DERIVED' and lbuff5[-1] == 'DERIVED':
                                    dict_member.update({'_type_id':int(lbuff5[1])})
                                    dict_member.update({'_is_derived':True})
                                    dict_member.update({'_forced_pad':True})
                                    dict_member.update({'_size':8})
                                    dict_member.update({'_stride':8})
                                if lbuff5[0] == 'CLASS' and lbuff5[-1] == 'CLASS':
                                    dict_member.update({'_size':8})
                                    dict_member.update({'_stride':8})
                        else:
                            if 'PROC_POINTER' in lbuff5 or lbuff5[0] == 'PROCEDURE':
                                nbytes = sizeof(c_void_p)
                                dict_member.update({'_is_proc': True})
                                dict_member.update({'_type'  : c_void_p})
                                dict_member.update({'_nbytes': nbytes})
                                dict_member.update({'_size'  : nbytes})
                                dict_member.update({'_stride': nbytes})
                                types2sizes.update({c_void_p : nbytes})
                        if dict_entry.get('_is_char', False):
                            if len(lbuff5) == 2:
                                if lbuff5[1].startswith("'") and lbuff5[1].endswith("'"):
                                    dict_entry.update({'_length':int(lbuff5[1].strip("'"))})
                        if dict_member.get('_is_char', False):
                            if sbuff5.strip() == 'DEFERRED_CL':
                                dict_member.update({'_is_deferred_char':True})
                                dict_member.update({'_stride':8})
                                dict_member.update({'_size':8})
                            if len(lbuff5) == 3:
                                if lbuff5[-1] == 'DEFERRED':
                                    dict_member.update({'_ndims':int(lbuff5[0])})
                                    dict_member.update({'_is_deferred':True})
                                    dict_member.update({'_forced_pad':True})
                                    dict_member.update({'_is_deferred_char': True})
                                    dict_member.setdefault('_stride', 8)
                                    dict_member.setdefault('_size', 8)
                            if dict_member.get('_is_deferred', False):
                                if lbuff5[-1] == 'POINTER':
                                    dict_member.update({'_is_pointer':True})
                            if not dict_member.get('_is_deferred_char', False):
                                if len(sbuff5.split("'")[0].split()) == 2:
                                    try:
                                        if int(int(lbuff5[1])) != dict_member.get('_length', -1):
                                            print(f'\n >>> WARNING: Character length {name_member} goes wrong!')
                                        dict_member.update({'_value':sbuff5.split("'")[-2][:dict_member['_length']]})
                                    except:
                                        pass
                        else:
                            if len(lbuff5):
                                if lbuff5[-1] == 'DEFERRED':
                                    dict_member.update({'_is_deferred':True})
                                    dict_member.update({'_ndims':int(lbuff5[0])})
                                if lbuff5[-1] == 'POINTER':
                                    dict_member.update({'_is_pointer':True})
                        if dict_member.get('_is_pending', False):
                            offset, offset_done = _recalcOffset(dict_member, sizes, offset, padding)
                            if offset_done and '_offset' not in dict_member:
                                raise RuntimeError(f' >>> DL_F2PY ERROR: _offset missing for {entry}%{kk if "kk" in locals() else name_member}')
                        sbuff5 = ''
                        fp.write('\n                ⟫')
                    elif depth == 5:
                        if is_internal:
                            sbuff6 = ''
                            fp.write('\n                    >')
                            continue
                        sbuff5 = ''
                        lbuff6 = sbuff6.split()
#                        self.__printSbuff(sbuff6, sbuff=6, depth=5, entry='arr03_of_int', debug=debug)
                        if dict_entry.get('_is_const', False) and dict_entry.get('_ndims', 0):
                            dt = lbuff4
                            hexadecimal = sbuff6.split()[-1].strip("'")
                            value = self.hexadecimal2number(hexadecimal, dict_entry['_type'])
                            dict_entry['_lbuff'] += [ value ]
                        if dict_entry.get('_no_member', False):
                            dict_member = {}
                        if dict_member.get('_ndims', 0):
                            if len(lbuff6) == 2:
                                if lbuff6[0].startswith("'") and lbuff6[0].endswith("'"):
                                    if lbuff6[1].startswith("'") and lbuff6[1].endswith("'"):
                                        shape = (int(lbuff6[0].strip("'")),int(lbuff6[1].strip("'")))
                                        if shape != dict_member['_dim']:
                                            print(f'\n >>> WARNING: The explicit shape of array {name_member} goes wrong!')
                                else:
                                    if lbuff6[1].startswith("'") and lbuff6[1].endswith("'"):
                                        dict_member.setdefault('_lbuff', [])
                                        dict_member['_lbuff'] += [ int(lbuff6[1].strip("'")) ]
                                        if not len(dict_member['_lbuff'])%2:
                                            dict_member.update({'_dim':tuple(dict_member['_lbuff'][i*2+1]-dict_member['_lbuff'][i*2]+1 for i in range(len(dict_member['_lbuff'][::2])))})
                                            if len(dict_member['_dim']) == dict_member.get('_ndims', 0):
                                                dict_member.pop('_lbuff', None)
                            if dict_member.get('_is_pending', False):
                                offset, offset_done = _recalcOffset(dict_member, sizes, offset, padding)
                        fp.write('\n                    >')
                        sbuff6 = ''
                    elif depth == 6:
                        if is_internal or dict_entry.get('_no_member', False) or 'dict_member' not in locals():
                            sbuff7 = ''
                            fp.write('\n                        ”')
                            continue
                        sbuff6 = ''
                        lbuff7 = sbuff7.split()
                        if dict_member.get('_is_char'):
                            if len(lbuff7) == 8 and lbuff7[-1].startswith("'") and lbuff7[-1].endswith("'"):
                                dict_member.update({'_length': int(lbuff7[-1].strip("'"))})
                                dict_member.update({'_size'  : int(lbuff7[-1].strip("'"))})
                                if dict_member.get('_is_pending', False):
                                    offset, offset_done = _recalcOffset(dict_member, sizes, offset, padding)
                                    if offset_done and '_offset' not in dict_member:
                                        raise RuntimeError(f' >>> DL_F2PY ERROR: _offset missing for pending char: {entry}%{name_member}')
                        fp.write('\n                        ”')
                        sbuff7 = ''
                    else:
                        pass
                else:
                    fp.write(c)
                if c not in ['(', ')']:
                    if depth == 0:
                        sbuff0 += c
                    elif depth == 1:
                        sbuff1 += c
                    elif depth == 2:
                        sbuff2 += c
                    elif depth == 3:
                        sbuff3 += c
                    elif depth == 4:
                        sbuff4 += c
                    elif depth == 5:
                        sbuff5 += c
                    elif depth == 6:
                        sbuff6 += c
                    else:
                        sbuff7 += c
        for k, v in _instances_derived_types.items():
            if is_internal:
                continue
            if type(v) is int:
                tmp = indices2entries.get(v, None)
                if tmp:
                    _instances_derived_types.update({k:tmp})
        self._instances_derived_types.update({basename:_instances_derived_types})
        derived_types = {}
        for k, v in dbuff.items():
            if type(v) is dict:
                if v.get('_is_derived', False):
                    derived_types.update({k:v})
        self._derived_types.update({basename:derived_types})
        def _findDerivedTypes(self, _type, _padding):
            _size_type = types2sizes[_type]
            for x in fortran_internals:
                if entry.startswith(x):
                    return 0
            if _type == '_tb_procedure':
                return 0
            for _k, _dict_subtype in dbuff[_type].items():
                if _k == '_tb_procedures':
                    continue
                if type(_dict_subtype) is dict:
                    _subtype = ids2types.get(_dict_subtype.get('_type_id', -1), _dict_subtype.get('_type', ''))
                    _dict_subtype.update({'_type':_subtype})
                    if _subtype in dbuff:
                        if _dict_subtype['_type'] != _subtype:
                            _dict_subtype.update({'_type_id':_dict_subtype['_type']})
                        _dict_subtype['_type'] = _subtype
                        if '_dim' in _dict_subtype:
                            size_array = prod(_dict_subtype['_dim'])
                            ndims = _dict_subtype['_ndims']
                        else:
                            size_array = 1
                            ndims = 1
                        if debug > 1:
                            print(f' >>>     {_clr._R}Old size{_clr.CLR_} of type {_clr._bC}{_clr._type}{_clr.CLR_}:', _size_type)
                        if _subtype == _type:
                            if '_ndims' in _dict_subtype:
                                if _dict_subtype.get('_is_deferred', False):
                                    _size_subtype = 40 + _dict_subtype['_ndims']*24 
                            else:
                                if _dict_subtype.get('_is_pointer', False):
                                    _size_subtype = 8
                        else:
                            _was_subtype_padded = _dict_subtype.get('_is_padded', False)
                            _size_subtype = _findDerivedTypes(self, _subtype, _was_subtype_padded)
                            if _dict_subtype.get('_is_padded', False) and _size_subtype == 4:
                                _dict_subtype.update({'_is_padded':False})
                                _size_type -= 4
                                types2sizes[_type] -= 4
                            _dict_subtype.update({'_size_chunk':_size_subtype})
                        if _dict_subtype.get('_is_deferred', False):
                            size_array = 0
                        if _dict_subtype.get('_is_pointer', False) and not '_ndims' in _dict_subtype:
                            _size_subtype = 8
                            size_array = 0
                        if _size_subtype:
                            _size_type += (_size_subtype-8)*size_array
                        if debug > 1:
                            print(f' >>>     {_clr._G}New size{_clr.CLR_} of type {_clr._bC}{_clr._type}{_clr.CLR_}:', _size_type)
            residue = _size_type%8
            if _size_type == 4:
                _padding = False
            if _padding and residue and 8 > residue:
                _size_type += 8 - residue
                if debug > 1:
                    print(f' >>>     Padded {_clr._bC}{_clr._type}{_clr.CLR_} by {8-residue} bytes to final size {_clr._bM}{_clr._size_type}{_clr.CLR_}')
            dbuff[_type].update({'_size_chunk':_size_type})
            return _size_type
        for k, v in dbuff.items():
            if type(v) is dict:
                tp = ids2types.get(types2ids.get('_type', v.get('_type_id', -1)), '')
                if tp:
                    v.update({'_type':tp})
                    size_new = _findDerivedTypes(self, tp, padding)
                    v.update({'_size_chunk':size_new})
                size_derived = 0
                accumulation = 0
                was_derived = False
                if v.get('_is_derived', False) and not k.startswith('__'):
                    for kk, vv in v.items():
                        if kk == '_tb_procedures':
                            continue
                        if type(vv) is dict:
                            tp = ids2types.get(types2ids.get('_type', vv.get('_type_id', -1)), '')
                            if tp:
                                vv.update({'_type':tp})
                            if debug > 1:
                                print(f' >>> {_clr._R}Old offset{_clr.CLR_} of {_clr._C}{k}{_clr.CLR_}%{_clr._Y}{kk}{_clr.CLR_}: {_clr._M}{vv["_offset"]}{_clr.CLR_}', f'(accumulation = {accumulation})')
                            offset_assumed = vv['_offset'] + accumulation - vv.get('_padded', 0)
                            if vv['_type'] == '_tb_procedure':
                                residue = 0
                            else:
                                stride = 8
                                residue = (offset_assumed)%stride
                            if vv.get('_is_padded') and residue:
                                if debug > 1:
                                    print(f' >>> {_clr._bM}Repadded{_clr.CLR_} {_clr._bY}"{kk}"{_clr.CLR_} by 4 because {size_derived}+{vv["_offset"]} has residue {residue}')
                                    print(f' >>> otherwise offset would be {offset_assumed} (stride = {vv["_stride"]})')
                                pad = abs(vv.get('_stride', 8) - residue)
                                vv['_offset'] += accumulation - vv.get('_padded', 0) + pad
                                accumulation += - vv.get('_padded', 0) + pad
                                if pad:
                                    vv.update({'_is_padded':True})
                                    vv.update({'_padded':pad})
                                else:
                                    vv.update({'_is_padded':False})
                                    vv.pop('padded', 0)
                            else:
                                if not vv.get('_forced_pad', False):
                                    vv['_offset'] += accumulation - vv.get('_padded', 0)
                                    accumulation -= vv.get('_padded', 0)
                                    vv.update({'_is_padded':False})
                                    vv.pop('padded', 0)
                                else:
                                    vv['_offset'] += accumulation
                            if debug > 1:
                                print(f' >>> {_clr._bG}New offset{_clr.CLR_} of {_clr._bC}{k}{_clr.CLR_}%{_clr._bY}{kk}{_clr.CLR_}: {_clr._bM}{vv["_offset"]}{_clr.CLR_}', f'(accumulation = {accumulation})')
                            if tp and kk != '_def_init':
                                vv.update({'_type':tp})
                                if '_dim' in vv:
                                    size_array = prod(vv['_dim'])
                                    ndims = vv['_ndims']
                                else:
                                    size_array = 1
                                    ndims = 1
                                if tp in dbuff:
                                    size_new = _findDerivedTypes(self, tp, padding)
                                    vv.update({'_size_chunk':size_new})
                                else:
                                    continue
                                if k == tp:
                                    size_new = types2sizes[tp]
                                size_derived = size_new*size_array
                                if vv.get('_is_deferred', False):
                                    ndims = vv['_ndims']
                                    size_derived = 0
                                    size_array = 0
                                if vv.get('_is_pointer', False) and not '_ndims' in vv:
                                    size_derived = 0
                                    size_array = 0
                                accumulation += size_derived
                                accumulation -= 8*size_array
                                if debug > 1:
                                    print(f' >>> {_clr._bG}New size{_clr.CLR_} of {_clr._bY}{k}{_clr.CLR_}.{_clr._bC}{tp}{_clr.CLR_}: {_clr._bM}{size_new}{_clr.CLR_} (size_derived = {size_derived}, accumulation = {accumulation})\n')
                                was_derived = True
                            else:
                                size_derived = 0
                                was_derived = False
        if debug:
            _printSummary(dbuff)
        dbuff.update({'_filename':modpath})
        object.__setattr__(self.modules, dbuff['_name'], dbuff)
        return dbuff
    def __printSbuff(self, s, sbuff=-1, depth=-1, entry='', debug=False):
        ccode = f'{_clr._bE}'
        if depth == 0:
            ccode = f'{_clr._W}'
        elif depth == 1:
            ccode = f'{_clr._R}'
        elif depth == 2:
            ccode = f'{_clr._Y}'
        elif depth == 3:
            ccode = f'{_clr._bY}'
        elif depth == 4:
            ccode = f'{_clr._G}'
        elif depth == 5:
            ccode = f'{_clr._C}'
        elif depth == 6:
            ccode = f'{_clr._B}'
        elif depth == 7:
            ccode = f'{_clr._M}'
        if entry == self.__entry or not entry:
            if debug:
                if sbuff < depth:
                    print(f' >>> DEBUG: entering {sbuff} -> {depth}, sbuff{sbuff} = {ccode}{s}{_clr.CLR_}')
                else:
                    print(f' >>> DEBUG: leaving {sbuff} -> {depth}, sbuff{sbuff} = {ccode}{s}{_clr.CLR_}')
    @property
    def derived_types(self):
         return list(self._derived_types.keys())
    @staticmethod
    def hexadecimal2number(sbuff, tp, return_ctype=False):
        from numpy import float16, float32, float64, int32, int64
        ctypes2ptypes = { c_int   : int32,
                          c_long  : int64,
                          c_float : float32,
                          c_double: float64,
                          c_bool  : lambda x:bool(int(x)),
                        }
        try:
            mbuff, ebuff = sbuff.split('@')
        except:
            return ctypes2ptypes[tp](sbuff)
        sign = ctypes2ptypes[tp](1.0)
        if mbuff.strip().startswith('-'):
            sign = ctypes2ptypes[tp](-1.0)
        ebuff = int(ebuff)
        point = mbuff.index('.')
        mbuff = mbuff[point+1:]
        mbuff = mbuff.ljust(ebuff, '0')
        hexadecimal = mbuff[:ebuff]+'.'+mbuff[ebuff:]
        try:
            result = ctypes2ptypes[tp](int(hexadecimal.split('.')[0], 16))
        except ValueError:
            result = ctypes2ptypes[tp](0.0)
        mantissa = hexadecimal.split('.')[1]
        for i, d in enumerate(mantissa):
            diviser = ctypes2ptypes[tp](16.0)**ctypes2ptypes[tp](i+1)
            result += ctypes2ptypes[tp](int(d,16))/diviser
        result *= sign
        return result
        if return_ctype:
            return tp(result)
        else:
            return tp(result).value
setters = {
            c_float : DL_DL.setFloat,
            c_double: DL_DL.setDouble,
            c_int   : DL_DL.setInt,
            c_long  : DL_DL.setLong,
            c_bool  : DL_DL.setBool,
            str     : DL_DL.setChar,
          }



#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
from struct import pack
def double2bytearray(d):
    return bytearray(pack('d', d))
def double2hexarray(d):
    ba = double2bytearray(d)
    return [ '0x%02x'%b for b in ba ]
def double2intarray(d):
    ha = double2hexarray(d)
    return [ int(h, base=16) for h in ha ]
def float2bytearray(f):
    return bytearray(pack('f', f))
def float2hexarray(f):
    ba = float2bytearray(f)
    return [ '0x%02x'%b for b in ba ]
def float2intarray(f):
    ha = float2hexarray(f)
    return [ int(h, base=16) for h in ha ]
def bool2bytearray(b):
    return bytearray(pack('?', b))
def bool2hexarray(b):
    ba = bool2bytearray(b)
    return [ '0x%02x'%b for b in ba ]
def bool2intarray(b):
    ha = bool2hexarray(b)
    return [ int(h, base=16) for h in ha ]
def int2bytearray(i):
    return bytearray(pack('i', i))
def int2hexarray(i):
    ba = int2bytearray(i)
    return [ '0x%02x'%b for b in ba ]
def int2intarray(i):
    ha = int2hexarray(i)
    return [ int(h, base=16) for h in ha ]
def long2bytearray(l):
    return bytearray(pack('l', l))
def long2hexarray(l):
    ba = long2bytearray(l)
    return [ '0x%02x'%b for b in ba ]
def long2intarray(l):
    ha = long2hexarray(l)
    return [ int(h, base=16) for h in ha ]
def char2bytearray(c):
    return bytearray(pack('c', c))
def char2hexarray(c):
    ba = char2bytearray(c)
    return [ '0x%02x'%b for b in ba ]
def char2intarray(c):
    ha = char2hexarray(c)
    return [ int(h, base=16) for h in ha ]



#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
__author__ = 'You Lu <you.lu@ukri.stfc.org>'
def __getEntry(arg, default):
    from . import iterutils
    if not arg:
        try:
            return iterutils.getFlattened((default,))
        except:
            return arg
    else:
        return iterutils.getFlattened((arg,))
def init(obj=None, inherit=(), init=(), priorities=(), inheritfrom=None):
    def __init(proc):
        from functools import wraps
        from .         import iterutils
        def __getObj():
            return obj
        @wraps(proc)
        def __wrapper(*args, **kwargs):
            obj = __getObj()
            if not obj:
                obj = args[0]
            def __getInheritfrom():
                if inheritfrom:
                    return iterutils.getFlattened((inheritfrom,))
                else:
                    return inheritfrom
            inheritAttrs(obj,
                         __getEntry(inherit, obj._inherit),
                         inheritfrom=__getInheritfrom())
            initAttrs(obj,
                      __getEntry(init      , obj._init),
                      __getEntry(priorities, obj._priorities))
            result = proc(*args, **kwargs)
            return result
        return __wrapper
    return __init
def inheritAttrs(obj, inherit, inheritfrom=None):
    from . import iterutils
    inheritSeqAttrs(obj,
                    inherit,
                    inheritfrom=inheritfrom)
    inheritDictAttrs(obj,
                     obj._init,
                     inheritfrom=inheritfrom)
def initAttrs(obj, init, priorities):
    initSeqAttrs(obj, init)
    initDictAttrs(obj, init, priorities=obj._priorities )
def inheritSeqAttrs(obj, attrs, inheritfrom=None):
    from . import iterutils, miscutils
    if not inheritfrom:
        inheritfrom = miscutils.getBaseClasses(obj.__class__)
    attrs = iterutils.getFlattened(['_inherit', '_init', list(attrs)])
    for baseCls in inheritfrom:
        it = iter(attrs)
        for i, attr in enumerate(it):
            try: 
                setattr(obj, attr, iterutils.getFlattened((getattr(baseCls,attr),)+(getattr(obj,attr),)))
            except:
                pass
def inheritDictAttrs(obj, attrs, inheritfrom=None):
    from .    import miscutils
    from copy import deepcopy
    if not inheritfrom:
        inheritfrom = miscutils.getBaseClasses(obj.__class__)
    for baseCls in inheritfrom:
        for attr in attrs:
            try:
                items = getattr(baseCls, attr).items()
                for key, val in items:
                    getattr(obj, attr).setdefault(key, deepcopy(val))
            except:
                continue
def initDictAttrs(obj, attrs, priorities=None):
    from copy import deepcopy
    items = getMergedItems(obj, attrs, order=priorities)
    for key, val in items:
        if key in getattr(obj, '_properties', []):
            addProperty(obj, key)
            continue
        try:
            setattr(obj, key, deepcopy(val))
            obj.__dict__.update()
        except (AttributeError, TypeError):
            setattr(obj, '_'+key, deepcopy(val))
        except:
            raise
def initSeqAttrs(obj, seqs):
    from . import iterutils
    for seq in seqs:
        val = getattr(obj, seq)
        if type(val) is dict:
            continue
        try:
            setattr(obj, seq, iterutils.getFlattened((val,)))
        except:
            setattr(obj, '_'+seq, iterutils.getFlattened((val,)))
def getMergedItems(obj, dicts, order=None):
    from . import dictutils
    items = []
    for attr in dicts:
        try:
            items += list(getattr(obj, attr).items())
        except:
            continue
    keys = dictutils.getSortedKeys(dict(items), order=order)
    return [ (key, dict(items)[key]) for key in keys ]
def checkType(obj, cls, info="", chksubcls=False):
    from numpy import ndarray, random, float64
    from .     import nputils
    if isinstance(obj, cls):
        if isinstance(obj, str):
            return obj
        if cls == list:
            if all(isinstance(i, type(obj[0])) for i in obj):
                return nputils.ndarray(obj)
            else:
                return obj
        if isinstance(obj, tuple):
            return ndarray(shape=obj, buffer=random.rand(obj[0]*obj[1]), dtype=float64)
        else:
            return obj
    elif cls is type(None):
        return obj
    elif isinstance(obj, cls.__bases__):
        return obj
    else:
        print(f"\n >>> DL_PY2F ERROR: Type {cls} expected for keyword {info} but {type(obj)} was given. Exiting...\n", flush=True)
        raise TypeError
assignByType = checkType
def getBaseType(obj):
    if hasattr(obj.__class__, '__base__'):
         try:
             return getBaseType(obj.__class__.__base__())
         except TypeError:
             return obj.__class__.__base__
    else:
         return obj.__class__
def getType(obj, list_modules=[], ndarrayislist=True):
    from numpy  import ndarray, float64, float32, int64, int32
    from types  import ModuleType, FunctionType
    if isinstance(obj, ndarray) and ndarrayislist:
        return list
    elif isinstance(obj, ndarray) and not ndarrayislist:
        return ndarray
    elif isinstance(obj, list):
        return list
    elif isinstance(obj, str):
        return str
    elif isinstance(obj, bytes):
        return bytes
    elif isinstance(obj, float64) or isinstance(obj, float64):
        return float
    elif isinstance(obj, int64) or isinstance(obj, int32):
        return int
    elif isinstance(obj, ModuleType):
        return ModuleType
    elif isinstance(obj, FunctionType):
        return FunctionType
    elif type(obj).__module__ in list_modules:
        return object
    else:
        return type(obj)
def getDeepCopy(obj):
    from copy  import deepcopy
    from numpy import copy as numpy_copy
    from numpy import ndarray, recarray
    from .     import nputils
    if isinstance(obj, nputils.RecArray):
        copied = recarray.copy(obj)
    elif isinstance(obj, ndarray):
        copied = numpy_copy(obj)
    else:
        copied = deepcopy(obj)
        if hasattr(obj, '__dict__'):
            for (name, attr) in [ (name, getattr(obj, name)) for name in obj.__dict__ if name not in {'_b_weakref_'} ]:
                setattr(copied, name, getDeepCopy(attr))
    return copied
def printTraceback(path):
    import os
    import sys
    import traceback
    print('\n >>> Failed to load module', path, flush=True)
    var = traceback.format_exc()
    print(var, flush=True)
    sys.exit(999)
def getModuleList(obj):
    from pkgutil import walk_packages, extend_path
    res = []
    def _recur(_path, _prefix): 
        for _loader, _name, _ispkg in walk_packages(_path, _prefix, onerror=printTraceback):
            pass
        try:
            for _loader, _name, _ispkg in walk_packages(_path, _prefix, onerror=printTraceback):
                pass
        except:
            print(" >>> ERROR: There is a bug found in module %s"%_name, flush=True)
            print(" >>> Use command `python3` to run your script for detailed reasons (requiring environment variables such as $PYTHONPATH)", flush=True)
            return
        for _loader, _name, _ispkg in walk_packages(_path, _prefix, onerror=printTraceback):
            _newpath = extend_path(_path, _name)
            _newlist = []
            try:
                _newlist.append(_newpath[1])
                _recur(_newlist, _name+'.')
            except:
                pass
            res.append(_name)
    _recur(obj.__path__, obj.__name__+'.')
    return set(res)
def getAttrType(obj, attr, list_modules=[]):
    try:
        return getType(getattr(obj, attr, list_modules=list_modules))
    except:
        return type(None)
def getAttr(obj, attrname, level=0):
    from . import strutils
    if not attrname:
        return obj
    try:
        return getattr(obj, attrname)
    except TypeError:
        return [ getAttr(obj, attr) for attr in attrname ]
    except AttributeError:
        return getAttr(getAttr(obj, 
                               strutils.split(attrname, '.')[0]),
                       strutils.split(attrname, '.', maxsplit=1)[1])
def type2dtype(obj, descr=False):
    from numpy import dtype, ndarray
    if descr:
        attr = 'descr'
    else:
        attr = 'str'
    selectcases = {
                    str    : lambda x: getattr(dtype(('S', len(obj)))    , attr),
                    int    : lambda x: getattr(dtype('int64')  , attr),
                    float  : lambda x: getattr(dtype('float64'), attr),
                    ndarray: lambda x: getattr(x.dtype, attr),
                    list   : lambda x: _getComplexDtype(x),
                    tuple  : lambda x: _getComplexDtype(x)
                  }
    def _getComplexDtype(array_seq):
        return [ ('', type2dtype(arr, descr=descr), (arr.shape[1],)) if arr.ndim>1 else
                 ('', type2dtype(arr, descr=descr),) for arr in array_seq ]
    return selectcases[type(obj)](obj)
def obj2dict(obj, list_modules=[]):
    from numpy import ndarray
    from .     import miscutils, nputils
    dictbuff = {}
    selectcases = {
                    bytes     : bytes.decode,
                    list      : miscutils.doNothing,
                    ndarray   : nputils.tolist,
                    str       : str.strip,
                    type(None): lambda x: None,
                    int       : int,
                    float     : float,
                    object    : obj2dict
                  }
    for key, val in obj._kwargs.items():
        dictbuff.update({key:selectcases[getType(getattr(obj,key,list_modules=list_modules), ndarrayislist=False)](getattr(obj, key))})
    return dictbuff
def setAttr(obj, attr, val, debug=0):
    from .         import nputils
    from .strutils import getRightmost
    if not attr:
        try:
            nputils.setField(obj, val, debug=debug)
            return
        except:
            return
    try:
        setattr(obj, attr, val)
    except TypeError:
        for key in attr:
            attrname = getRightmost(key)
            destin = getAttr(obj, getRightmost(key, pos=1))
            if hasattr(getattr(destin, attrname), 'size'):
                destin.resize(val.size)
                setattr(destin, attrname, val[attrname])
            else:
                raise AttributeError
def addProperty(obj, attr):
    mangled = '_'+attr
    setattr(obj, attr, property(lambda self: getattr(self, mangled),
                                lambda self, val: setattr(self, mangled, val)))



#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
__author__ = 'You Lu <you.lu@ukri.stfc.org>'
from . import ctypeutils, dictutils, fileutils, iterutils, miscutils, modutils, nputils, objutils, strutils
__all__ = 'ctypeutils', 'dictutils', 'fileutils', 'iterutils', 'miscutils', 'modutils', 'nputils', 'objutils', 'strutils'



import subprocess
from pathlib import Path

from setuptools import setup, find_packages
from setuptools.command.build_py import build_py as _build_py

ROOT = Path(__file__).parent.resolve()
PKG_DIR = ROOT / 'dl_py2f'
SRC_DIR = ROOT / 'dl_py2f'
BUILD_LIB_DIR = ROOT / 'build'   # arbitrary name for CMake build dir


class build_py(_build_py):
    '''
    Custom build_py that first builds libdl_py2f.so via CMake
    and copies it into dl_py2f/ so it ends up inside the wheel.
    '''

    def run(self):
        self.build_dl_py2f_lib()
        super().run()

    def build_dl_py2f_lib(self):
        # Ensure build directory exists
        BUILD_LIB_DIR.mkdir(exist_ok=True)

        # ---- 1. Configure CMake ----
        # Adapt these commands to match your existing build.
        # For example, if your current script does:
        #   cmake -B build -S . -DCMAKE_BUILD_TYPE=Release
        #   cmake --build build --target dl_py2f
        # you can do the same here:
        cmake_config_cmd = [
            'cmake',
            '-S', str(SRC_DIR),
            '-B', str(BUILD_LIB_DIR),
            '-DCMAKE_BUILD_TYPE=Release',
        ]
        subprocess.run(cmake_config_cmd, check=True)

        # ---- 2. Build the library ----
        cmake_build_cmd = [
            'cmake',
            '--build', str(BUILD_LIB_DIR),
            '--target', 'dl_py2f',
            '--config', 'Release',
        ]
        subprocess.run(cmake_build_cmd, check=True)

        # ---- 3. Verify build outputs exist in package directory ----
        # CMakeLists.txt outputs .so and .mod directly to dl_py2f/
        built_lib = PKG_DIR / 'libdl_py2f.so'
        if not built_lib.exists():
            raise RuntimeError(f'Could not find {built_lib} after CMake build')
        print(f'Built library: {built_lib}')

        built_mods = list(PKG_DIR.glob('*.mod'))
        if not built_mods:
            raise RuntimeError('No .mod file could be found in package directory')
        for mod in built_mods:
            print(f'Built module: {mod}')

setup(name='dl_py2f',
      version='1.1.9',
      description='DL_PY2F: A library for Python-Fortran interoperability',
      author='You Lu',
      author_email='you.lu@stfc.ac.uk',
      url='https://github.com/stfc/dl_py2f',
      license='GNU Lesser General Public License v3.0',
      packages=find_packages(),
      include_package_data=True,
      package_data={'dl_py2f': ['*.so', '*.mod']},
      python_requires='>=3.9',
      install_requires=['numpy>=1.26'],
      cmdclass={'build_py': build_py})




#  Copyright (C) 2025 The authors of DL_PY2F
#
#  This file is part of DL_PY2F.
#
#  DL_PY2F is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  DL_PY2F is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with DL_PY2F. If not, see
#  <https://www.gnu.org/licenses/>.

# you.lu@stfc.ac.uk
__author__ = 'You Lu <you.lu@ukri.stfc.org>'
def str2num(s, delimiter=' '):
    try:
        try:
            return int(s)
        except ValueError:
            return float(s)
    except ValueError:
         from numpy    import fromstring
         from warnings import catch_warnings, simplefilter
         with catch_warnings(record=True) as w:
             simplefilter('error')
             try:
                 return fromstring(s, sep=delimiter)
             except:
                 return []
def importAsModule(s):
    from importlib import import_module
    try:
        module = import_module(s)
    except OSError:
        print("\n DL_PY2F ERROR: Library", s, "is not correctly compiled/linked.\n")
        quit(999)
    except ImportError:
        module = s
    return module
def getJoined(s, bar, newline=True):
    s += bar
    if newline:
        s += '\n'
    return s
class Str(bytearray):
    def __init__(self, val, *args):
        self = bytearray(val, 'ascii')
    def __iadd__(self, *args):
        try:
            self.extend((args[0]+'\n').encode())
        except:
            if args[0][1] in [ 'nonewline', 'nonl' ]:
                self.extend(args[0][0].encode())
            else:
                self.extend((args[0][0]+'\n').encode())
        return self
    def clear(self):
        del self[:]
    @property
    def s(self):
        return self.decode()
def getRightmost(s, separator='.', pos=0, tolerant=False):
    try:
        return s.rsplit(separator, pos+1)[-pos-1]
    except:
        if tolerant:
            return getRightmost(s, separator=separator, pos=pos-1, tolerant=tolerant)
        else:
            return ''
def split(s, separator='.', maxsplit=-1, tolerant=False):
    if s.split(separator, maxsplit=maxsplit)[0] == s and not tolerant:
        return ['', '']
    return s.split(separator, maxsplit=maxsplit)
def bytes2str(b):
    try:
        return b.decode()
    except:
        return str(b)
def str2bytes(s):
    if issubclass(type(s), bytes):
        return s
    try:
        return s.encode()
    except:
        return str(s).encode()
def inOrEq(s, ref):
    result = False
    try:
        s = s.decode().lower()
    except:
        s = s.lower()
    try:
        if s == ref.lower():
            result = True
    except AttributeError:
        refs = [ r.lower() for r in ref ]
        if s in refs:
            result = True
    return result
def isStr(s):
    from numpy import str_, string_
    if type(s) in [ str, str_, string_ ]:
        return True
    else:
        return False
